"""
Provides base class for static objects. Static objects are those
objects that do not update continuously based on the telemetry
feed.
"""
from PIL import ImageFont, Image, ImageDraw
from moviepy.video.io.bindings import PIL_to_npimage


class StaticBase:
    """
    Defines base class for static objects, including default object
    return methods.
    """
    _row_colors = [
        (192, 192, 192, 255),
        (255, 255, 255, 255),
    ]

    def __init__(self, data, size=None, **kwargs):
        self._data = data
        self._options = kwargs
        self._columns = list()
        self._size = size

    def sort_data(self, call):
        self._data = sorted(self._data, key=call)

    @property
    def row_colors(self):
        return self._row_colors

    @row_colors.setter
    def row_colors(self, value):
        self._row_colors = value

    def add_column(self, attribute, heading=None, *,
                   align='left', formatter=None, formatter_args=None,
                   colspan=1):
        self._columns.append(DisplayColumn(
            attribute,
            heading,
            align=align,
            formatter=formatter,
            formatter_args=formatter_args,
            colspan=colspan))

    def add_lookup(self, attribute, lookup, default, heading=None, *,
                   align='left', formatter=None, formatter_args=None):
        self._columns.append(DisplayColumn(
            attribute,
            heading,
            lookup,
            default,
            align=align,
            formatter=formatter,
            formatter_args=formatter_args))

    def to_frame(self):
        return PIL_to_npimage(self._make_material().convert('RGBA'))

    def _make_material(self):
        DisplayLine.reset()

        #  If data exists, create a heading.
        try:
            heading_color = tuple(self._options['heading_color'])
            heading_font_color = tuple(
                self._options['heading_font_color'])
            try:
                heading_font = ImageFont.truetype(
                    self._options['heading_font'],
                    self._options['heading_font_size'])
            except (AttributeError, OSError):
                heading_font = ImageFont.load_default()

            try:
                series_logo = Image.open(self._options['series_logo'])
            except (AttributeError, KeyError, OSError):
                series_logo = None

            heading_text = self._options['heading_text'] if self._options['heading_text'] is not None else ""
            subheading_text = self._options['subheading_text'] if self._options['subheading_text'] is not None else ""
        except KeyError:
            heading_color = None
            heading_font_color = None
            heading_font = None
            heading_text = ""
            subheading_text = ""
            heading = False
            series_logo = None
        else:
            heading = True

        #  If provided, use a font.
        try:
            try:
                font = ImageFont.truetype(
                    self._options['font'],
                    self._options['font_size'])
            except (AttributeError, OSError):
                font = ImageFont.load_default()
            font_color = tuple(self._options['font_color'])
        except KeyError:
            font = ImageFont.load_default()
            font_color = (0, 0, 0)

        #  If set, use external and internal margins.
        try:
            margin = self._options['margin']
        except KeyError:
            margin = 2*font.getsize("A")[1]

        try:
            column_margin = self._options['column_margin']
        except KeyError:
            column_margin = 1*font.getsize("A")[1]

        #  If set, use limit on number of lines to show.
        try:
            starting_grid = \
                self._data[:self._options['result_lines']]
        except KeyError:
            starting_grid = self._data

        #  If set, use a backdrop.
        try:
            backdrop = Image.open(self._options['backdrop'])

            if self._size is not None:
                backdrop = backdrop.resize(self._size)
                backdrop_size = self._size
            else:
                backdrop_size = backdrop.size
        except (AttributeError, KeyError, IOError):
            backdrop = None
            backdrop_size = None

        #  If set, use a logo on the backdrop.
        try:
            logo = Image.open(self._options['logo'])
            logo_size = (
                self._options['logo_width'],
                self._options['logo_height'])
        except (AttributeError, KeyError, IOError):
            logo = None
            logo_size = None

        #  Build main data material.
        column_widths = list()
        display_lines = list()

        if any([column.heading for column in self._columns]):
            line = DisplayLine(self._columns, None, True)
            column_widths = line.column_widths(font, column_widths)
            display_lines.append(line)

        for grid in starting_grid:
            line = DisplayLine(self._columns, grid)
            column_widths = line.column_widths(font, column_widths)
            display_lines.append(line)

        row_height = int(font.getsize("A")[1]*2.5)
        material_width = sum(column_widths)\
            + (len(column_widths)-1)*column_margin
        material_width += 2*margin

        #  Build heading, if applicable.
        heading_height = 0
        heading_material = None
        if heading:
            heading_block_height = sum([
                heading_font.getsize("A")[1],
                font.getsize("A")[1]])
            heading_height = int(heading_block_height*2.5)

            series_logo_width = 0
            series_logo_height = 0
            if series_logo is not None:
                series_logo.thumbnail((material_width, heading_height))
                series_logo_width, series_logo_height = series_logo.size

            material_width = max([
                (
                    2*margin
                    + heading_font.getsize(str(heading_text))[0]
                    + series_logo_width),
                (
                    2*margin
                    + font.getsize(str(subheading_text))[0]
                    + series_logo_height),
                material_width])

            heading_material = Image.new(
                'RGBA',
                (material_width, heading_height),
                heading_color)
            if series_logo is not None:
                text_x_position = material_width-series_logo_width
                y_position = 0
                heading_material.paste(
                    series_logo,
                    (text_x_position, y_position))

            draw = ImageDraw.Draw(heading_material)

            x_position = margin
            y_position = int((heading_height-heading_block_height)/2)

            draw.text(
                (x_position, y_position),
                heading_text,
                fill=heading_font_color,
                font=heading_font)
            y_position += heading_font.getsize("A")[1]
            draw.text(
                (margin, y_position),
                subheading_text,
                fill=heading_font_color,
                font=font)

        material_height = sum([
            heading_height,
            row_height*len(starting_grid)])
        if any([column.heading for column in self._columns]):
            material_height += row_height

        material = Image.new(
            'RGBA',
            (material_width, material_height))

        #  Write heading, if applicable.
        if heading:
            material.paste(heading_material, (0, 0), heading_material)

        y_position = heading_height
        for index, line in enumerate(display_lines):
            row_material = Image.new(
                'RGBA',
                (material_width, row_height),
                self._row_color(index))
            draw = ImageDraw.Draw(row_material)

            text_x_position = margin
            text_y_position = int((row_height-font.getsize("A")[1])/2)

            iline = iter(line)
            for display_text, column_width, align, col_span in iline:
                if isinstance(display_text, str):
                    if align == 'center' \
                            and line.heading_line \
                            and col_span > 1:
                        span = col_span
                        while span > 1:
                            _, next_column_width, _, _ = next(iline)
                            column_width += \
                                next_column_width \
                                + column_margin
                            span -= 1

                    if align == 'center':
                        text_width = font.getsize(display_text)[0]
                        draw_x_position = text_x_position + int(
                            (column_width-text_width)/2)
                    else:
                        draw_x_position = text_x_position
                    draw.text(
                        (draw_x_position, text_y_position),
                        display_text,
                        fill=font_color,
                        font=font)
                else:
                    row_material.paste(
                        display_text,
                        (text_x_position, text_y_position),
                        display_text)

                text_x_position += column_width + column_margin
            material.paste(row_material, (0, y_position), row_material)
            y_position += row_height

        if backdrop is not None:
            backdrop_width, backdrop_height = backdrop_size

            #  Add logo if needed.
            if logo is not None:
                logo = logo.resize(logo_size)
                logo_width, logo_height = logo_size
                text_x_position = backdrop_width-logo_width
                y_position = backdrop_height-logo_height
                backdrop.paste(
                    logo,
                    (text_x_position, y_position),
                    logo)

            if material_width > backdrop_width \
                    or material_height > backdrop_height:
                material.thumbnail(backdrop_size)
                material_width, material_height = material.size

            text_x_position = int((backdrop_width-material_width)/2)
            y_position = int((backdrop_height-material_height)/2)
            backdrop.paste(
                material,
                (text_x_position, y_position),
                material)
            material = backdrop

        return material

    def _row_color(self, index):
        return self._row_colors[index % len(self._row_colors)]

    @staticmethod
    def _row_width(widths, margin, internal_margin):
        return sum(widths) \
               + 2 * margin \
               + (len(widths) - 1) * internal_margin


class DisplayColumn:
    """
    Defines a single column in the display.
    """
    _point_structure = list()

    def __init__(self, attribute, heading=None, lookup=None,
                 default=None, *, align='left', formatter=None,
                 formatter_args=None, colspan=1):
        self._heading = heading
        self._attribute = attribute
        self._lookup = lookup
        self._default = default
        self._align = align
        self._formatter = formatter
        self._formatter_args = formatter_args
        self._colspan = colspan

    @property
    def align(self):
        return self._align

    @property
    def attribute(self):
        return self._attribute

    @property
    def colspan(self):
        return self._colspan

    @property
    def default(self):
        return self._default

    @property
    def formatter(self):
        return self._formatter

    @property
    def formatter_args(self):
        return self._formatter_args

    @property
    def heading(self):
        return self._heading

    @property
    def lookup(self):
        return self._lookup

    @property
    def point_structure(self):
        return self._point_structure

    @point_structure.setter
    def point_structure(self, value):
        self._point_structure = value


class DisplayLine:
    """
    Defines a single line in the display.
    """
    _column_widths = list()

    def __init__(self, columns, data, make_headings=False):
        self._line_data = list()
        self._col_spans = list()
        self._align = list()
        self._heading_line = make_headings

        for column in columns:
            self._align.append(column.align)
            self._col_spans.append(column.colspan)
            try:
                if make_headings:
                    text_value = str(column.heading)
                else:
                    text_value = column.lookup[
                        getattr(data, column.attribute)]
            except (KeyError, TypeError):
                if column.default is None:
                    text_value = getattr(data, column.attribute)
                else:
                    text_value = column.default

            if column.formatter is None or make_headings:
                self._line_data.append(str(text_value))
            elif column.formatter_args is None:
                self._line_data.append(column.formatter(text_value))
            else:
                self._line_data.append(column.formatter(
                    text_value,
                    **column.formatter_args))

    def __iter__(self):
        line_data = iter(self._line_data)
        column_widths = iter(self._column_widths)
        col_spans = iter(self._col_spans)
        align = iter(self._align)
        while True:
            yield (
                next(line_data),
                next(column_widths),
                next(align),
                next(col_spans))

    @property
    def heading_line(self):
        return self._heading_line

    def column_widths(self, font, column_widths):
        for index, data in enumerate(self._line_data):
            if self._heading_line and self._col_spans[index] > 1:
                self._column_widths.append(0)
            elif isinstance(data, str):
                try:
                    self._column_widths[index] = max(
                        [column_widths[index], font.getsize(data)[0]])
                except IndexError:
                    self._column_widths.append(font.getsize(data)[0])
            else:
                try:
                    self._column_widths[index] = max(
                        [column_widths[index], data.size[0]])
                except IndexError:
                    self._column_widths.append(data.size[0])

        return self._column_widths

    @classmethod
    def reset(cls):
        cls._lookups = list()
        cls._defaults = list()
        cls._column_widths = list()



"""
Tests StaticBase.py
"""
import unittest
from unittest.mock import patch, sentinel

import numpy
from PIL import Image, ImageFont

from replayenhancer.StaticBase import StaticBase
from replayenhancer.StartingGridEntry import StartingGridEntry


class TestStaticBase(unittest.TestCase):
    """
    Tests against the StaticBase object.
    """
    test_data = [
        StartingGridEntry(1, 0, "Kobernulf Monnur"),
        StartingGridEntry(2, 1, "Testy McTest")
    ]

    def test_init(self):
        instance = StaticBase(self.test_data)
        expected_value = StaticBase
        self.assertIsInstance(instance, expected_value)

    @patch('replayenhancer.StaticBase.PIL_to_npimage', autospec=True)
    def test_method_to_frame(self, mock_mpy):
        mock_mpy.return_value = numpy.array(sentinel.test)
        instance = StaticBase(self.test_data)
        expected_value = numpy.ndarray
        self.assertIsInstance(instance.to_frame(), expected_value)

    @patch('replayenhancer.StaticBase.ImageFont.truetype',
           autospec=True)
    @patch('replayenhancer.StaticBase.PIL_to_npimage', autospec=True)
    def test_method_to_frame_with_header(self, mock_mpy, mock_font):
        mock_font.return_value = ImageFont.load_default()
        configuration = {
            'heading_color': [255, 0, 0],
            'heading_font_color': [255, 255, 255],
            'heading_font': sentinel.font,
            'heading_font_size': sentinel.font_size,
            'heading_text': "Lorem Ipsum",
            'subheading_text': "For one, for all"
        }
        mock_mpy.return_value = numpy.array(sentinel.test)
        instance = StaticBase(self.test_data, **configuration)
        expected_value = numpy.ndarray
        self.assertIsInstance(instance.to_frame(), expected_value)

    @patch('replayenhancer.StaticBase.ImageFont.truetype',
           autospec=True)
    @patch('replayenhancer.StaticBase.PIL_to_npimage', autospec=True)
    def test_method_to_frame_with_incomplete_header(self,
                                                    mock_mpy,
                                                    mock_font):
        mock_font.return_value = ImageFont.load_default()
        configuration = {
            'heading_color': [255, 0, 0],
            'heading_font_color': [255, 255, 255],
            'heading_text': "Lorem Ipsum",
            'subheading_text': "For one, for all"
        }
        mock_mpy.return_value = numpy.array(sentinel.test)
        instance = StaticBase(self.test_data, **configuration)
        expected_value = numpy.ndarray
        self.assertIsInstance(instance.to_frame(), expected_value)

    @patch('replayenhancer.StaticBase.ImageFont.truetype',
           autospec=True)
    @patch('replayenhancer.StaticBase.PIL_to_npimage', autospec=True)
    def test_method_to_frame_with_bad_header(self, mock_mpy, mock_font):
        mock_font.side_effect = OSError
        configuration = {
            'heading_color': [255, 0, 0],
            'heading_font_color': [255, 255, 255],
            'heading_text': "Lorem Ipsum",
            'subheading_text': "For one, for all"
        }
        mock_mpy.return_value = numpy.array(sentinel.test)
        instance = StaticBase(self.test_data, **configuration)
        expected_value = numpy.ndarray
        self.assertIsInstance(instance.to_frame(), expected_value)

    @patch('replayenhancer.StaticBase.ImageFont.truetype',
           autospec=True)
    @patch('replayenhancer.StaticBase.PIL_to_npimage', autospec=True)
    def test_method_to_frame_with_font(self, mock_mpy, mock_font):
        mock_font.return_value = ImageFont.load_default()
        configuration = {
            'font': sentinel.font,
            'font_size': sentinel.font_size,
            'font_color': [0, 0, 0]
        }
        mock_mpy.return_value = numpy.array(sentinel.test)
        instance = StaticBase(self.test_data, **configuration)
        expected_value = numpy.ndarray
        self.assertIsInstance(instance.to_frame(), expected_value)

    @patch('replayenhancer.StaticBase.ImageFont.truetype',
           autospec=True)
    @patch('replayenhancer.StaticBase.PIL_to_npimage', autospec=True)
    def test_method_to_frame_with_incomplete_font(self,
                                                  mock_mpy,
                                                  mock_font):
        mock_font.return_value = ImageFont.load_default()
        configuration = {
            'font': sentinel.font
        }
        mock_mpy.return_value = numpy.array(sentinel.test)
        instance = StaticBase(self.test_data, **configuration)
        expected_value = numpy.ndarray
        self.assertIsInstance(instance.to_frame(), expected_value)

    @patch('replayenhancer.StaticBase.ImageFont.truetype',
           autospec=True)
    @patch('replayenhancer.StaticBase.PIL_to_npimage', autospec=True)
    def test_method_to_frame_with_bad_font(self, mock_mpy, mock_font):
        mock_font.side_effect = OSError()
        configuration = {
            'font': sentinel.font
        }
        mock_mpy.return_value = numpy.array(sentinel.test)
        instance = StaticBase(self.test_data, **configuration)
        expected_value = numpy.ndarray
        self.assertIsInstance(instance.to_frame(), expected_value)

    @patch('replayenhancer.StaticBase.PIL_to_npimage', autospec=True)
    def test_method_to_frame_with_margin(self, mock_mpy):
        configuration = {
            'margin': 20
        }
        mock_mpy.return_value = numpy.array(sentinel.test)
        instance = StaticBase(self.test_data, **configuration)
        expected_value = numpy.ndarray
        self.assertIsInstance(instance.to_frame(), expected_value)

    @patch('replayenhancer.StaticBase.PIL_to_npimage', autospec=True)
    def test_method_to_frame_with_column_margin(self, mock_mpy):
        configuration = {
            'column_margin': 10
        }
        mock_mpy.return_value = numpy.array(sentinel.test)
        instance = StaticBase(self.test_data, **configuration)
        expected_value = numpy.ndarray
        self.assertIsInstance(instance.to_frame(), expected_value)

    @patch('replayenhancer.StaticBase.Image.open', autospec=True)
    @patch('replayenhancer.StaticBase.PIL_to_npimage', autospec=True)
    def test_method_to_frame_with_backdrop(self, mock_mpy, mock_img):
        mock_img.return_value = Image.new('L', (800, 800))
        configuration = {
            'backdrop': sentinel.backdrop
        }
        mock_mpy.return_value = numpy.array(sentinel.test)
        instance = StaticBase(self.test_data, **configuration)
        expected_value = numpy.ndarray
        self.assertIsInstance(instance.to_frame(), expected_value)

    @patch('replayenhancer.StaticBase.Image.open', autospec=True)
    @patch('replayenhancer.StaticBase.PIL_to_npimage', autospec=True)
    def test_method_to_frame_with_backdrop_and_logo(self,
                                                    mock_mpy,
                                                    mock_img):
        mock_img.return_value = Image.new('L', (800, 800))
        configuration = {
            'backdrop': sentinel.backdrop,
            'logo': sentinel.logo,
            'logo_height': 150,
            'logo_width': 150
        }
        mock_mpy.return_value = numpy.array(sentinel.test)
        instance = StaticBase(self.test_data, **configuration)
        expected_value = numpy.ndarray
        self.assertIsInstance(instance.to_frame(), expected_value)

    @patch('replayenhancer.StaticBase.ImageFont.truetype',
           autospec=True)
    @patch('replayenhancer.StaticBase.PIL_to_npimage', autospec=True)
    def test_method_to_frame_full_configuration(self,
                                                mock_mpy,
                                                mock_font):
        mock_font.return_value = ImageFont.load_default()
        configuration = {
            'heading_color': [255, 0, 0],
            'heading_font_color': [255, 255, 255],
            'heading_font': sentinel.font,
            'heading_font_size': sentinel.font_size,
            'heading_text': "Lorem Ipsum",
            'subheading_text': "For one, for all",
            'font': sentinel.font,
            'font_size': sentinel.font_size,
            'margin': 20,
            'column_margin': 10
        }
        mock_mpy.return_value = numpy.array(sentinel.test)
        instance = StaticBase(self.test_data, **configuration)
        expected_value = numpy.ndarray
        self.assertIsInstance(instance.to_frame(), expected_value)

if __name__ == '__main__':
    unittest.main()



"""
Tests GTStandings.py
"""

import unittest
from unittest.mock import MagicMock, patch

from replayenhancer.GTStandings import Animation, GTStandings, PitStopFlyout, StandingLine


class TestPitStopFlyout(unittest.TestCase):
    """Unit tests for PitStopFlyout object.

    """
    def test_init(self):
        race_data = MagicMock(spec='replayenhancer.RaceData.RaceData', autospec=True)
        driver = MagicMock(spec='replayenhancer.RaceData.Driver', autospec=True)
        font = MagicMock(spec='PIL.ImageFont', autospec=True)
        size = (100, 100)
        location = (100, 1, 100)

        instance = PitStopFlyout(race_data, driver, font, size, location)
        expected_result = PitStopFlyout
        self.assertIsInstance(instance, expected_result)

    def test_init_margin(self):
        race_data = MagicMock(spec='replayenhancer.RaceData.RaceData', autospec=True)
        driver = MagicMock(spec='replayenhancer.RaceData.Driver', autospec=True)
        font = MagicMock(spec='PIL.ImageFont', autospec=True)
        size = (100, 100)
        location = (100, 1, 100)

        instance = PitStopFlyout(race_data, driver, font, size, location, margin=42)
        expected_result = PitStopFlyout
        self.assertIsInstance(instance, expected_result)

    def test_init_ups(self):
        race_data = MagicMock(spec='replayenhancer.RaceData.RaceData', autospec=True)
        driver = MagicMock(spec='replayenhancer.RaceData.Driver', autospec=True)
        font = MagicMock(spec='PIL.ImageFont', autospec=True)
        size = (100, 100)
        location = (100, 1, 100)

        instance = PitStopFlyout(race_data, driver, font, size, location, ups=60)
        expected_result = PitStopFlyout
        self.assertIsInstance(instance, expected_result)

    def test_method_close_flyout(self):
        race_data = MagicMock(spec='replayenhancer.RaceData.RaceData', autospec=True)
        driver = MagicMock(spec='replayenhancer.RaceData.Driver', autospec=True)
        font = MagicMock(spec='PIL.ImageFont', autospec=True)
        size = (100, 100)
        location = (100, 1, 100)

        instance = PitStopFlyout(race_data, driver, font, size, location)
        instance.close_flyout()
        expected_result = Animation
        self.assertIsInstance(instance.animations[-1], expected_result)

    def test_method_close_flyout_set_persist(self):
        race_data = MagicMock(spec='replayenhancer.RaceData.RaceData', autospec=True)
        driver = MagicMock(spec='replayenhancer.RaceData.Driver', autospec=True)
        font = MagicMock(spec='PIL.ImageFont', autospec=True)
        size = (100, 100)
        location = (100, 1, 100)

        instance = PitStopFlyout(race_data, driver, font, size, location)
        instance.close_flyout()
        self.assertFalse(instance.persist)

    def test_method_close_flyout_set_is_closing(self):
        race_data = MagicMock(spec='replayenhancer.RaceData.RaceData', autospec=True)
        driver = MagicMock(spec='replayenhancer.RaceData.Driver', autospec=True)
        font = MagicMock(spec='PIL.ImageFont', autospec=True)
        size = (100, 100)
        location = (100, 1, 100)

        instance = PitStopFlyout(race_data, driver, font, size, location)
        instance.close_flyout()
        self.assertTrue(instance.is_closing)

    @patch('PIL.ImageDraw.Draw', autospec=True)
    @patch('PIL.Image', autospec=True)
    def test_method_to_frame(self, mock_image, mock_draw):
        race_data = MagicMock(spec='replayenhancer.RaceData.RaceData', autospec=True)
        driver = MagicMock(spec='replayenhancer.RaceData.Driver', autospec=True)
        font = MagicMock(spec='PIL.ImageFont', autospec=True)
        size = (100, 100)
        location = (100, 1, 100)

        mock_image.new = MagicMock(spec='numpy.ndarray', autospec=True)
        font.getsize = MagicMock(return_value=(20, 20))
        mock_draw.text = MagicMock(spec='numpy.ndarray')

        instance = PitStopFlyout(race_data, driver, font, size, location)
        from PIL.Image import Image as expected_value
        self.assertIsInstance(instance.to_frame(), expected_value)

    def test_method_update_not_stopped(self):
        race_data = MagicMock(spec='replayenhancer.RaceData.RaceData', autospec=True)
        driver = MagicMock(spec='replayenhancer.RaceData.Driver', autospec=True)
        font = MagicMock(spec='PIL.ImageFont', autospec=True)
        size = (100, 100)
        location = (100, 1, 100)

        instance = PitStopFlyout(race_data, driver, font, size, location)
        for time in range(16):
            instance.update((time*10, 1, time*10), time)

    def test_method_update_stopped_no_time(self):
        race_data = MagicMock(spec='replayenhancer.RaceData.RaceData', autospec=True)
        driver = MagicMock(spec='replayenhancer.RaceData.Driver', autospec=True)
        font = MagicMock(spec='PIL.ImageFont', autospec=True)
        size = (100, 100)
        location = (100, 1, 100)

        instance = PitStopFlyout(race_data, driver, font, size, location)
        for _ in range(16):
            instance.update(location)

    def test_method_update_stopped_ascending_time(self):
        race_data = MagicMock(spec='replayenhancer.RaceData.RaceData', autospec=True)
        driver = MagicMock(spec='replayenhancer.RaceData.Driver', autospec=True)
        font = MagicMock(spec='PIL.ImageFont', autospec=True)
        size = (100, 100)
        location = (100, 1, 100)

        instance = PitStopFlyout(race_data, driver, font, size, location)
        for time in range(16):
            instance.update(location, time)

    def test_method_update_stopped_reset_time(self):
        race_data = MagicMock(spec='replayenhancer.RaceData.RaceData', autospec=True)
        driver = MagicMock(spec='replayenhancer.RaceData.Driver', autospec=True)
        font = MagicMock(spec='PIL.ImageFont', autospec=True)
        size = (100, 100)
        location = (100, 1, 100)

        instance = PitStopFlyout(race_data, driver, font, size, location)
        for time in range(8):
            instance.update(location, time + 8)
        for time in range(8):
            instance.update(location, time)

    def test_property_text_color_pit_in(self):
        race_data = MagicMock(spec='replayenhancer.RaceData.RaceData',
                              autospec=True)
        driver = MagicMock(spec='replayenhancer.RaceData.Driver', autospec=True)
        font = MagicMock(spec='PIL.ImageFont', autospec=True)
        size = (100, 100)
        location = (100, 1, 100)

        instance = PitStopFlyout(race_data, driver, font, size, location)
        expected_result = (255, 0, 0, 255)
        self.assertTupleEqual(instance.text_color, expected_result)

    def test_property_text_color_pit_out(self):
        race_data = MagicMock(spec='replayenhancer.RaceData.RaceData', autospec=True)
        driver = MagicMock(spec='replayenhancer.RaceData.Driver', autospec=True)
        font = MagicMock(spec='PIL.ImageFont', autospec=True)
        size = (100, 100)
        location = (0, 1, 0)

        instance = PitStopFlyout(race_data, driver, font, size, location)
        for time in range(16):
            instance.update(location, time)

        expected_result = (255, 255, 255, 255)
        self.assertTupleEqual(instance.text_color, expected_result)

    def test_field_is_closing(self):
        race_data = MagicMock(spec='replayenhancer.RaceData.RaceData', autospec=True)
        driver = MagicMock(spec='replayenhancer.RaceData.Driver', autospec=True)
        font = MagicMock(spec='PIL.ImageFont', autospec=True)
        size = (100, 100)
        location = (0, 1, 0)

        instance = PitStopFlyout(race_data, driver, font, size, location)
        self.assertFalse(instance.is_closing)

    def test_field_is_closing_set_true(self):
        race_data = MagicMock(spec='replayenhancer.RaceData.RaceData', autospec=True)
        driver = MagicMock(spec='replayenhancer.RaceData.Driver', autospec=True)
        font = MagicMock(spec='PIL.ImageFont', autospec=True)
        size = (100, 100)
        location = (0, 1, 0)

        instance = PitStopFlyout(race_data, driver, font, size, location)
        instance.is_closing = True
        self.assertTrue(instance.is_closing)

    def test_field_persist(self):
        race_data = MagicMock(spec='replayenhancer.RaceData.RaceData', autospec=True)
        driver = MagicMock(spec='replayenhancer.RaceData.Driver', autospec=True)
        font = MagicMock(spec='PIL.ImageFont', autospec=True)
        size = (100, 100)
        location = (0, 1, 0)

        instance = PitStopFlyout(race_data, driver, font, size, location)
        self.assertTrue(instance.persist)

    def test_field_persist_set_false(self):
        race_data = MagicMock(spec='replayenhancer.RaceData.RaceData', autospec=True)
        driver = MagicMock(spec='replayenhancer.RaceData.Driver', autospec=True)
        font = MagicMock(spec='PIL.ImageFont', autospec=True)
        size = (100, 100)
        location = (0, 1, 0)

        instance = PitStopFlyout(race_data, driver, font, size, location)
        instance.persist = False
        self.assertFalse(instance.persist)




@unittest.skip("Need to rewrite based on new GTStandings module.")
class TestGTStandings(unittest.TestCase):
    """
    Unit tests for a GT Standings object.
    """

    def test_init(self):
        instance = GTStandings()
        expected_value = GTStandings
        self.assertIsInstance(instance, expected_value)

    def test_property_default_clip_t(self):
        instance = GTStandings()
        expected_value = 0
        self.assertEqual(instance.clip_t, expected_value)

    def test_property_clip_t(self):
        instance = GTStandings(11.0)
        expected_value = 11.0
        self.assertEqual(instance.clip_t, expected_value)

    def test_property_default_ups(self):
        instance = GTStandings()
        expected_value = 30
        self.assertEqual(instance.ups, expected_value)

    def test_property_ups(self):
        instance = GTStandings(ups=10)
        expected_value = 10
        self.assertEqual(instance.ups, expected_value)

    # def test_method_to_frame(self):
    #     instance = GTStandings()
    #     expected_value = PIL_to_npimage(
    #         Image.new('RGBA', (100, 100)).convert('RGB'))
    #     assert_array_equal(instance.to_frame(), expected_value)

    def test_method_update(self):
        instance = GTStandings()
        expected_value = None
        self.assertEqual(instance.update(), expected_value)

@unittest.skip("Need to rewrite based on new GTStandings module.")
class TestStandingLine(unittest.TestCase):
    """
    Unit tests for the StandingLine object.
    """

    @patch('PIL.ImageFont.ImageFont')
    @patch('replayenhancer.RETelemetryDataPacket.REParticipantInfo')
    def test_init(self, mock_driver, mock_font):
        instance = StandingLine(mock_driver, mock_font, (100, 100))
        expected_value = StandingLine
        self.assertIsInstance(instance, expected_value)

    @patch('PIL.ImageFont.ImageFont')
    @patch('replayenhancer.RETelemetryDataPacket.REParticipantInfo')
    def test_name_color_not_viewed(self, mock_driver, mock_font):
        instance = StandingLine(mock_driver, mock_font, (100, 100))
        instance._viewed = False
        expected_value = instance._name_color
        self.assertTupleEqual(instance.name_color, expected_value)

    @patch('PIL.ImageFont.ImageFont')
    @patch('replayenhancer.RETelemetryDataPacket.REParticipantInfo')
    def test_name_text_color_not_viewed(self, mock_driver, mock_font):
        instance = StandingLine(mock_driver, mock_font, (100, 100))
        instance._viewed = False
        expected_value = instance._name_text_color
        self.assertTupleEqual(instance.name_text_color, expected_value)

    @patch('PIL.ImageFont.ImageFont')
    @patch('replayenhancer.RETelemetryDataPacket.REParticipantInfo')
    def test_name_color_viewed(self, mock_driver, mock_font):
        instance = StandingLine(mock_driver, mock_font, (100, 100))
        instance._viewed = True
        expected_value = instance._viewed_name_color
        self.assertTupleEqual(instance.name_color, expected_value)

    @patch('PIL.ImageFont.ImageFont')
    @patch('replayenhancer.RETelemetryDataPacket.REParticipantInfo')
    def test_name_text_color_viewed(self, mock_driver, mock_font):
        instance = StandingLine(mock_driver, mock_font, (100, 100))
        instance._viewed = True
        expected_value = instance._viewed_name_text_color
        self.assertTupleEqual(instance.name_text_color, expected_value)

    @patch('PIL.ImageFont.ImageFont')
    @patch('replayenhancer.RETelemetryDataPacket.REParticipantInfo')
    def test_property_position_color_not_viewed(self, mock_driver, mock_font):
        instance = StandingLine(mock_driver, mock_font, (100, 100))
        instance._viewed = False
        expected_value = instance._position_color
        self.assertTupleEqual(instance.position_color, expected_value)

    @patch('PIL.ImageFont.ImageFont')
    @patch('replayenhancer.RETelemetryDataPacket.REParticipantInfo')
    def test_property_position_text_color_not_viewed(self, mock_driver, mock_font):
        instance = StandingLine(mock_driver, mock_font, (100, 100))
        instance._viewed = False
        expected_value = instance._position_text_color
        self.assertTupleEqual(
            instance.position_text_color,
            expected_value)

    @patch('PIL.ImageFont.ImageFont')
    @patch('replayenhancer.RETelemetryDataPacket.REParticipantInfo')
    def test_property_position_color_viewed(self, mock_driver, mock_font):
        instance = StandingLine(mock_driver, mock_font, (100, 100))
        instance._viewed = True
        expected_value = instance._viewed_position_color
        self.assertTupleEqual(instance.position_color, expected_value)

    @patch('PIL.ImageFont.ImageFont', autospec=True)
    @patch(
        'replayenhancer.RETelemetryDataPacket.REParticipantInfo',
        autospec=True)
    def test_property_position_text_color_viewed(self, mock_driver, mock_font):
        instance = StandingLine(mock_driver, mock_font, (100, 100))
        instance._viewed = True
        expected_value = instance._viewed_position_text_color
        self.assertTupleEqual(
            instance.position_text_color,
            expected_value)

    # @patch(
    #     'replayenhancer.RETelemetryDataPacket.REParticipantInfo',
    #     autospec=True)
    # def test_method_render_not_viewed(self, mock_driver):
    #     font = ImageFont.load_default()
    #     mock_driver.race_position = 1
    #     mock_driver.name = "Kobernulf Monnur"
    #     instance = StandingLine(mock_driver, font, (100, 100))
    #     expected_value = Image.Image
    #     self.assertIsInstance(instance._render(), expected_value)
    #
    # @patch(
    #     'replayenhancer.RETelemetryDataPacket.REParticipantInfo',
    #     autospec=True)
    # def test_method_render_viewed(self, mock_driver):
    #     font = ImageFont.load_default()
    #     mock_driver.race_position = 1
    #     mock_driver.name = "Kobernulf Monnur"
    #     instance = StandingLine(mock_driver, font, (100, 100))
    #     instance._viewed = True
    #     expected_value = Image.Image
    #     self.assertIsInstance(instance._render(), expected_value)



"""
Tests StartingGridEntry.py
"""
from unittest import TestCase
from unittest.mock import sentinel

from replayenhancer.StartingGridEntry import StartingGridEntry


class TestStartingGridEntry(TestCase):
    """
    Tests StartingGridEntry object.
    """

    def test_init(self):
        instance = StartingGridEntry(sentinel.position,
                                     sentinel.driver_index,
                                     sentinel.driver_name)
        expected_result = StartingGridEntry
        self.assertIsInstance(instance, expected_result)

    def test_property_driver_index(self):
        instance = StartingGridEntry(sentinel.position,
                                     sentinel.driver_index,
                                     sentinel.driver_name)
        expected_result = sentinel.driver_index
        self.assertEqual(instance.driver_index, expected_result)

    def test_property_driver_name(self):
        instance = StartingGridEntry(sentinel.position,
                                     sentinel.driver_index,
                                     sentinel.driver_name)
        expected_result = sentinel.driver_name
        self.assertEqual(instance.driver_name, expected_result)

    def test_property_position(self):
        instance = StartingGridEntry(sentinel.position,
                                     sentinel.driver_index,
                                     sentinel.driver_name)
        expected_result = sentinel.position
        self.assertEqual(instance.position, expected_result)



"""
Captures UDP telemetry packets for use by the Project CARS Replay
Enhancer. Run on the network to which Project CARS is broadcasting

Writes the packets to a directory named "packetdata" with an
appended timestamp. Each packet is named "pdata" with an appended
sequence number.

Stop telemetry packet capture by hitting CTRL+C.
"""
import datetime
import os
import socket

def main():
    # Create a new UDP socket.
    SOCKET = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    SOCKET.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    # Bind the socket to the port
    SERVER_ADDRESS = ("", 5606)
    print("Starting listener on port {}".format(SERVER_ADDRESS[1]))
    SOCKET.bind(SERVER_ADDRESS)

    i = 0
    DIRECTORY = "packetdata-"+datetime.datetime.now().strftime(
        "%Y%m%d-%H%M%S")
    try:
        if not os.path.exists(DIRECTORY):
            os.makedirs(DIRECTORY)
        while True:
            DATA, _ = SOCKET.recvfrom(65565)
            print("Writing packet {}".format(i))
            FILE = open('./'+DIRECTORY+'/pdata'+str(i), 'wb')
            FILE.write(DATA)
            FILE.close()
            i += 1

    except KeyboardInterrupt:
        print("Closing listener on port {}".format(SERVER_ADDRESS[1]))

    finally:
        if i == 0:
            os.rmdir(DIRECTORY)

if __name__ == '__main__':
    main()



from replayenhancer.RETelemetryDataPacket import RETelemetryDataPacket
from glob import glob
from natsort import natsorted

with open('outputs/times.txt', 'w+') as f:
    for packet in natsorted(
                    glob('assets/race1-descriptor/pdata*')):
        with open(packet, 'rb') as packet_file:
            packet_data = packet_file.read()

        if len(packet_data) == 1367:
            RETDP = RETelemetryDataPacket(packet_data)
            time = str(RETDP.current_time) + " | " + str(RETDP.data_hash) + "\n"
            f.write(str(time))



"""
Provides base class for the UDP Packets output by Project CARS.
"""

import abc
from collections import deque
from struct import unpack


class Packet(metaclass=abc.ABCMeta):
    """
    Defines base Packet class for UDP Packets output by Project CARS.
    """
    @abc.abstractproperty
    def packet_type(self):
        """Define the packet type number of the packet."""

    @abc.abstractproperty
    def _packet_string(self):
        """Define the binary unpacking string for the packet."""

    def _test_packet_type(self, packet_type):
        """
        Tests the packet type against the defined packet type
        """
        try:
            packet_type = int(packet_type) & int('00000011', 2)
            if packet_type != self.packet_type:
                raise ValueError(
                    "Incorrect packet type detected. " +
                    "Packet type is {}. ".format(packet_type) +
                    "Packet type should be {}".format(
                        self.packet_type))
        except ValueError:
            raise
        else:
            return True

    def _unpack_data(self, packet_data):
        """
        Unpacks the binary data according to the string that
        represents the data structure.
        """
        return deque(unpack(self._packet_string, packet_data))

    def __repr__(self):
        return self.__str__()

    @abc.abstractmethod
    def __str__(self):
        """String representation of the packet."""



"""
Provides classes for default static cards.
"""
from PIL import Image, ImageDraw, ImageFont

from replayenhancer.StaticBase import StaticBase


class RaceResults(StaticBase):
    """
    Defines a class for a default Race Results title card.

    This card, by default, has the following columns:
        - Pos.: Finish position.
        - Driver: Driver name.
        - Team: Driver team (if applicable, blank else).
        - Car: Driver car.
        - Laps: Driver laps completed.
        - Stops: Pit stops (if greater than 0)
        - Time: Driver total race time.
        - Best Lap: Driver best lap.
        - Best S1: Driver best sector 1.
        - Best S2: Driver best sector 2.
        - Best S3: Driver best sector 3.
        - Points: Driver points earned.
    """
    def __init__(self, data, size=None, **kwargs):
        super().__init__(data, size=size, **kwargs)
        self.sort_data(lambda x: (-x.driver.laps_complete, x.driver.race_time))

        try:
            name_lookup = {
                k: v['display']
                for k, v in kwargs['participant_config'].items()}
        except KeyError:
            name_lookup = None

        try:
            car_lookup = {
                k: v['car']
                for k, v in kwargs['participant_config'].items()
                if v['car'] != ""}
            if len(car_lookup) == 0:
                car_lookup = None
        except KeyError:
            car_lookup = None

        try:
            team_lookup = {
                k: v['team']
                for k, v in kwargs['participant_config'].items()
                if v['team'] != ""}
            if len(team_lookup) == 0:
                team_lookup = None
        except KeyError:
            team_lookup = None

        try:
            point_structure = {
                k: v
                for k, v in enumerate(kwargs['point_structure'])}
            if len(point_structure) == 0 or not any(point_structure.values()):
                point_structure = None
        except KeyError:
            point_structure = None

        try:
            points_adjust = {
                k: v['points_adjust']
                for k, v in kwargs['participant_config'].items()
                if v['points_adjust'] != ""}
            if len(points_adjust) == 0:
                points_adjust = None
        except KeyError:
            points_adjust = None

        self.add_column('position', 'Pos.')

        if name_lookup is None:
            self.add_column('driver_name', 'Driver')
        else:
            self.add_lookup(
                'driver_name',
                name_lookup,
                'ERROR',
                'Driver')

        if team_lookup is not None:
            self.add_lookup('driver_name', team_lookup, '', 'Team')

        if car_lookup is not None:
            self.add_lookup('driver_name', car_lookup, '', 'Car')

        self.add_column('laps_complete', 'Laps', align='center')

        try:
            result_lines = self._options['result_lines']
            stops = [x.driver.stops for x in self._data[:result_lines]]
        except KeyError:
            stops = [x.driver.stops for x in self._data]

        if any(stops):
            self.add_column('stops', 'Stops', align='center')

        self.add_column(
            'race_time',
            'Time',
            formatter=self.format_time,
            align='center')
        self.add_column(
            'best_lap',
            'Best Lap',
            formatter=self.format_time,
            align='center')
        self.add_column(
            'best_sector_1',
            'Best S1',
            formatter=self.format_time,
            align='center')
        self.add_column(
            'best_sector_2',
            'Best S2',
            formatter=self.format_time,
            align='center')
        self.add_column(
            'best_sector_3',
            'Best S3',
            formatter=self.format_time,
            align='center')

        if point_structure is not None:
            formatter_args = {
                'point_structure': point_structure,
                'points_adjust': points_adjust}
            self.add_column(
                'calc_points_data',
                'Points',
                formatter=self.calc_points,
                formatter_args=formatter_args,
                align='center')

    def calc_points(self, value, **kwargs):
        points = self.calc_points_float(value, **kwargs)
        return str(int(points // 1))

    def calc_points_float(self, value, **kwargs):
        driver_name, position, best_lap = value
        points = 0
        try:
            if best_lap == min(
                    [entry.best_lap for entry in self._data if entry.best_lap is not None]):
                points += kwargs['point_structure'][0]
            points += kwargs['point_structure'][position]
        except (KeyError, TypeError):
            points += 0

        if 'points_adjust' in kwargs \
                and kwargs['points_adjust'] is not None \
                and driver_name in kwargs['points_adjust']:
            adjust = kwargs['points_adjust'][driver_name]
            if adjust[0] == '+':
                points += float(adjust[1:])
            elif adjust[0] == '-':
                points -= float(adjust[1:])
            else:
                try:
                    points = float(adjust)
                except ValueError:
                    pass

        return points

    @staticmethod
    def format_time(seconds):
        """
        Converts seconds into seconds, minutes:seconds, or
        hours:minutes.seconds as appropriate.
        """
        try:
            minutes, seconds = divmod(float(seconds), 60)
            hours, minutes = divmod(minutes, 60)

            return_value = (int(hours), int(minutes), float(seconds))

            if hours:
                return "{0:d}:{1:0>2d}:{2:0>6.3f}".format(*return_value)
            elif minutes:
                return "{1:d}:{2:0>6.3f}".format(*return_value)
            else:
                return "{2:.3f}".format(*return_value)
        except (TypeError, ValueError):
            return ""


class StartingGrid(StaticBase):
    """
    Defines a class for a default Starting Grid title card.

    This card, by default, has the following columns:
        - Pos.: Starting position.
        - Driver: Driver name.
        - Team: Driver team (if applicable, blank else).
        - Car: Driver car.
        - Points: Driver series points (if applicable, blank else).
    """
    def __init__(self, data, size=None, **kwargs):
        super().__init__(data, size=size, **kwargs)

        try:
            name_lookup = {
                k: v['display']
                for k, v in kwargs['participant_config'].items()}
        except KeyError:
            name_lookup = None

        try:
            car_lookup = {
                k: v['car']
                for k, v in kwargs['participant_config'].items()
                if v['car'] != ""}
            if len(car_lookup) == 0:
                car_lookup = None
        except KeyError:
            car_lookup = None

        try:
            team_lookup = {
                k: v['team']
                for k, v in kwargs['participant_config'].items()
                if v['team'] != ""}
            if len(team_lookup) == 0:
                team_lookup = None
        except KeyError:
            team_lookup = None

        try:
            point_structure = {
                k: v
                for k, v in enumerate(kwargs['point_structure'])}
            if len(point_structure) == 0 or not any(point_structure.values()):
                point_structure = None
        except KeyError:
            point_structure = None

        if point_structure is not None:
            try:
                points_lookup = {
                    k: v['points']
                    for k, v in kwargs['participant_config'].items()}
                if not any(points_lookup.values()):
                    points_lookup = None
            except KeyError:
                points_lookup = None
        else:
            points_lookup = None

        self.add_column('position', 'Pos.')

        if name_lookup is None:
            self.add_column('driver_name', 'Driver')
        else:
            self.add_lookup(
                'driver_name',
                name_lookup,
                'ERROR',
                'Driver')

        if team_lookup is not None:
            self.add_lookup('driver_name', team_lookup, '', 'Team')

        if car_lookup is not None:
            self.add_lookup('driver_name', car_lookup, '', 'Car')

        if points_lookup is not None or point_structure is not None:
            self.add_lookup(
                'driver_name',
                points_lookup,
                0,
                'Points',
                align='center')


class SeriesStandings(RaceResults):
    """
    Defines a class for a default Series Standings title card.

    This card, by default, has the following columns:
        - Rank: Series rank.
        - Driver: Driver name.
        - Team: Driver team (if applicable, blank else).
        - Car: Driver car.
        - Points: Driver series points.
    """

    def __init__(self, data, size=None, **kwargs):
        super(RaceResults, self).__init__(data, size=size, **kwargs)

        try:
            name_lookup = {
                k: v['display']
                for k, v in kwargs['participant_config'].items()}

            if 'additional_participant_config' in kwargs:
                for name in kwargs['additional_participant_config'].keys():
                    name_lookup[name] = name
        except KeyError:
            name_lookup = None

        try:
            car_lookup = {
                k: v['car']
                for k, v in kwargs['participant_config'].items()
                if v['car'] != ""}
            if 'additional_participant_config' in kwargs:
                for name, values in kwargs['additional_participant_config'].items():
                    if values['car'] != "":
                        car_lookup[name] = values['car']
            if len(car_lookup) == 0:
                car_lookup = None
        except KeyError:
            car_lookup = None

        try:
            team_lookup = {
                k: v['team']
                for k, v in kwargs['participant_config'].items()
                if v['team'] != ""}
            if 'additional_participant_config' in kwargs:
                for name, values in kwargs['additional_participant_config'].items():
                    if values['team'] != "":
                        team_lookup[name] = values['team']
            if len(team_lookup) == 0:
                team_lookup = None
        except KeyError:
            team_lookup = None

        try:
            point_structure = {
                k: v
                for k, v in enumerate(kwargs['point_structure'])}
        except KeyError:
            point_structure = None

        try:
            points_lookup = {
                k: v['points']
                for k, v in kwargs['participant_config'].items()}
            if 'additional_participant_config' in kwargs:
                for name, values in kwargs['additional_participant_config'].items():
                    points_lookup[name] = values['points']
        except KeyError:
            points_lookup = None

        try:
            points_adjust = {
                k: v['points_adjust']
                for k, v in kwargs['participant_config'].items()
                if v['points_adjust'] != ""}
            if len(points_adjust) == 0:
                points_adjust = None
        except KeyError:
            points_adjust = None

        if 'additional_participant_config' in kwargs:
            for name in kwargs['additional_participant_config'].keys():
                self._data.append(AdditionalClassificationEntry(name))

        formatter_args = {'point_structure': point_structure,
                          'points_lookup': points_lookup,
                          'points_adjust': points_adjust}
        self.sort_data(
            lambda x: (
                -int(self.calc_series_points(
                    x.calc_points_data, **formatter_args)),
                x.driver_name))

        if ('hide_series_zeros' in kwargs) and (
                    kwargs['hide_series_zeros'] is True):
            self._data = [x for x in self._data if int(self.calc_series_points(
                x.calc_points_data, **formatter_args)) != 0]

        self.add_column(
            'calc_points_data',
            'Rank',
            formatter=self.calc_series_rank,
            formatter_args=formatter_args)

        if name_lookup is None:
            self.add_column('driver_name', 'Driver')
        else:
            self.add_lookup(
                'driver_name',
                name_lookup,
                'ERROR',
                'Driver')

        if team_lookup is not None:
            self.add_lookup('driver_name', team_lookup, '', 'Team')

        if car_lookup is not None:
            self.add_lookup('driver_name', car_lookup, '', 'Car')

        self.add_column(
            'calc_points_data',
            'Points',
            formatter=self.calc_series_points,
            formatter_args=formatter_args,
            align='center')

    def calc_series_points(self, value, **kwargs):
        driver_name, position, best_lap = value
        try:
            points = kwargs['points_lookup'][driver_name]
        except (KeyError, TypeError):
            points = 0

        points += int(self.calc_points_float(value, **kwargs))

        return str(points)

    def calc_series_rank(self, value, **kwargs):
        driver_name, position, best_lap = value
        ranks = dict()
        last_points = None
        last_rank = 0
        for entry in self._data:
            if last_points != int(
                    self.calc_series_points(
                        entry.calc_points_data,
                        **kwargs)):
                last_points = int(
                    self.calc_series_points(
                        entry.calc_points_data,
                        **kwargs))
                last_rank = len(ranks) + 1
            ranks[entry.driver_name] = last_rank

        return str(ranks[driver_name])


class SeriesChampion(SeriesStandings):
    """
    Defines a class for a default Series Champion title card.
    """
    def __init__(self, data, size=None, **kwargs):
        super(RaceResults, self).__init__(data, size=size, **kwargs)

        try:
            self._name_lookup = {
                k: v['display']
                for k, v in kwargs['participant_config'].items()}

            if 'additional_participant_config' in kwargs:
                for name in kwargs['additional_participant_config'].keys():
                    self._name_lookup[name] = name
        except KeyError:
            self._name_lookup = {
                entry.driver_name: entry.driver_name
                for entry in self._data}

        try:
            self._car_lookup = {
                k: v['car']
                for k, v in kwargs['participant_config'].items()
                if v['car'] != ""}
            if 'additional_participant_config' in kwargs:
                for name, values in kwargs['additional_participant_config'].items():
                    if values['car'] != "":
                        self._car_lookup[name] = values['car']
            if len(self._car_lookup) == 0:
                self._car_lookup = None
        except KeyError:
            self._car_lookup = None

        try:
            self._team_lookup = {
                k: v['team']
                for k, v in kwargs['participant_config'].items()
                if v['team'] != ""}
            if 'additional_participant_config' in kwargs:
                for name, values in kwargs['additional_participant_config'].items():
                    if values['team'] != "":
                        self._team_lookup[name] = values['team']
            if len(self._team_lookup) == 0:
                self._team_lookup = None
        except KeyError:
            self._team_lookup = None

        try:
            point_structure = {
                k: v
                for k, v in enumerate(kwargs['point_structure'])}
        except KeyError:
            point_structure = None

        try:
            points_lookup = {
                k: v['points']
                for k, v in kwargs['participant_config'].items()}
            if 'additional_participant_config' in kwargs:
                for name, values in kwargs['additional_participant_config'].items():
                    points_lookup[name] = values['points']
        except KeyError:
            points_lookup = None

        try:
            points_adjust = {
                k: v['points_adjust']
                for k, v in kwargs['participant_config'].items()
                if v['points_adjust'] != ""}
            if len(points_adjust) == 0:
                points_adjust = None
        except KeyError:
            points_adjust = None

        if 'additional_participant_config' in kwargs:
            for name in kwargs['additional_participant_config'].keys():
                self._data.append(AdditionalClassificationEntry(name))

        formatter_args = {'point_structure': point_structure,
                          'points_lookup': points_lookup,
                          'points_adjust': points_adjust}
        self._formatter_args = formatter_args

        self.sort_data(
            lambda x: (
                -int(self.calc_series_points(
                    x.calc_points_data, **formatter_args)),
                x.driver_name))

    def _make_material(self):
        #  If data exists, create a heading.
        try:
            heading_color = tuple(self._options['heading_color'])

            try:
                heading_font_color = tuple(
                    self._options['heading_font_color'])
            except KeyError:
                heading_font_color = (0, 0, 0)

            try:
                heading_font = ImageFont.truetype(
                    self._options['heading_font'],
                    self._options['heading_font_size'])
            except (KeyError, OSError):
                heading_font = ImageFont.load_default()

            try:
                series_logo = Image.open(self._options['series_logo'])
            except (KeyError, OSError):
                series_logo = None

            try:
                heading_text = self._options['heading_text']
            except KeyError:
                heading_text = None

        except KeyError:
            heading_color = None
            heading_font_color = (0, 0, 0)
            heading_font = ImageFont.load_default()
            heading_text = None
            heading = False
            series_logo = None

        else:
            heading = True

        #  If provided, use a font.
        try:
            try:
                font = ImageFont.truetype(
                    self._options['font'],
                    self._options['font_size'])
            except OSError:
                font = ImageFont.load_default()
            font_color = tuple(self._options['font_color'])
        except KeyError:
            font = ImageFont.load_default()
            font_color = (0, 0, 0)

        #  If set, use external and internal margins.
        try:
            margin = self._options['margin']
        except KeyError:
            margin = 6*font.getsize("A")[1]

        try:
            column_margin = self._options['column_margin']
        except KeyError:
            column_margin = 3*font.getsize("A")[1]

        #  If set, use a backdrop.
        try:
            backdrop = Image.open(self._options['backdrop'])

            if self._size is not None:
                backdrop = backdrop.resize(self._size)
                backdrop_size = self._size
            else:
                backdrop_size = backdrop.size

        except (AttributeError, KeyError, IOError):
            backdrop = None
            backdrop_size = None

        #  If set, use a logo on the backdrop.
        try:
            logo = Image.open(self._options['logo'])
            logo_size = (
                self._options['logo_width'],
                self._options['logo_height'])
        except (AttributeError, KeyError, IOError):
            logo = None
            logo_size = None

        champion_data = [
            entry for entry in self._data
            if int(self.calc_series_rank(
                entry.calc_points_data,
                **self._formatter_args)) <= 3]

        #  Build main data material
        text_width = 2 * margin
        text_height = 2 * margin
        for rank, entry in enumerate(champion_data, 1):
            if rank == 1:
                width, height = heading_font.getsize("Champion")
                text_width = max([text_width, width + 2 * margin])
                text_height += height

                width, height = heading_font.getsize(
                    self._name_lookup[entry.driver_name])
                text_width = max([text_width, width + 2 * margin])
                text_height += height
            else:
                width, height = font.getsize("Runner Up")
                text_width = max([text_width, width + 2 * margin])
                text_height += height

                width, height = font.getsize(
                    self._name_lookup[entry.driver_name])
                text_width = max([text_width, width + 2 * margin])
                text_height += height

            if self._team_lookup is not None:
                width, height = font.getsize(
                    self._team_lookup[entry.driver_name])
                text_width = max([
                    text_width,
                    width + column_margin + 2 * margin])
                text_height += height

            if self._car_lookup is not None:
                width, height = font.getsize(
                    self._car_lookup[entry.driver_name])
                text_width = max([
                    text_width,
                    width + column_margin + 2 * margin])
                text_height += height

            text_height += margin

        #  TODO: Parametrize the "big logo" width.
        material_width = 300+text_width

        #  Build heading, if applicable.
        heading_height = 0
        heading_material = None
        if heading:
            heading_height = heading_font.getsize(str(heading_text))[1]
            heading_height += 2*margin

            material_width = max([
                (
                    2*margin
                    + heading_font.getsize(str(heading_text))[0]),
                material_width])

            heading_material = Image.new(
                'RGBA',
                (material_width, heading_height),
                heading_color)

            draw = ImageDraw.Draw(heading_material)

            if heading_text is not None:
                draw.text(
                    (margin, margin),
                    heading_text,
                    fill=heading_font_color,
                    font=heading_font)

        material_height = sum([
            heading_height,
            max([300, text_height])
        ])

        #  TODO: Parametrize background color.
        material = Image.new(
            'RGBA',
            (material_width, material_height),
            (255, 255, 255, 255))

        #  Write heading, if applicable.
        if heading:
            material.paste(heading_material, (0, 0), heading_material)

        if series_logo is not None:
            series_logo.thumbnail((300, 300))
            material.paste(series_logo, (0, heading_height))

        y_position = heading_height+int((material_height-text_height)/2)
        x_position = 300+margin

        draw = ImageDraw.Draw(material)

        for rank, entry in enumerate(champion_data, 1):
            if rank == 1:
                draw.text(
                    (x_position, y_position),
                    "Champion",
                    fill=font_color,
                    font=heading_font)
                y_position += heading_font.getsize("A")[1]

                draw.text(
                    (x_position, y_position),
                    self._name_lookup[entry.driver_name],
                    fill=font_color,
                    font=heading_font)
                y_position += heading_font.getsize("A")[1]
                x_position += column_margin
            else:
                draw.text(
                    (x_position, y_position),
                    "Runner Up",
                    fill=font_color,
                    font=font)
                y_position += font.getsize("A")[1]

                draw.text(
                    (x_position, y_position),
                    self._name_lookup[entry.driver_name],
                    fill=font_color,
                    font=font)
                y_position += font.getsize("A")[1]
                x_position += column_margin

            if self._team_lookup is not None:
                draw.text(
                    (x_position, y_position),
                    self._team_lookup[entry.driver_name],
                    fill=font_color,
                    font=font)
                y_position += font.getsize("A")[1]

            if self._car_lookup is not None:
                draw.text(
                    (x_position, y_position),
                    self._car_lookup[entry.driver_name],
                    fill=font_color,
                    font=font)
                y_position += font.getsize("A")[1]

            y_position += margin
            x_position -= column_margin

        if backdrop is not None:
            backdrop_width, backdrop_height = backdrop_size

            #  Add logo if needed.
            if logo is not None:
                logo = logo.resize(logo_size)
                logo_width, logo_height = logo_size
                text_x_position = backdrop_width-logo_width
                y_position = backdrop_height-logo_height
                backdrop.paste(
                    logo,
                    (text_x_position, y_position),
                    logo)

            if material_width > backdrop_width \
                    or material_height > backdrop_height:
                material.thumbnail(backdrop_size)
                material_width, material_height = material.size

            text_x_position = int((backdrop_width-material_width)/2)
            y_position = int((backdrop_height-material_height)/2)
            backdrop.paste(
                material,
                (text_x_position, y_position),
                material)
            material = backdrop

        return material


class AdditionalClassificationEntry:
    def __init__(self, name):
        self._name = name

    @property
    def best_lap(self):
        return None

    @property
    def calc_points_data(self):
        return (self._name, None, None)

    @property
    def driver_name(self):
        return self._name



# import sys
# import os
# import runpy
#
# path = os.path.dirname(sys.modules[__name__].__file__)
# path = os.path.join(path, '..')
# sys.path.insert(0, path)
# runpy.run_module('replayenhancer', run_name='replayenhancer', alter_sys=True)

from replayenhancer import replayenhancer
if __name__ == '__main__':
    replayenhancer.main()


"""
Provides a class for a Race Results card with position change.
"""
from PIL import Image, ImageDraw, ImageFont

from replayenhancer.DefaultCards import RaceResults


class RaceResultsWithChange(RaceResults):
    """
    Defines a class for a Race Results card with position change.
    """
    def __init__(self, data, starting_grid, size=None, **kwargs):
        super().__init__(data, size=size, **kwargs)

        position_lookup = dict()
        for entry in data:
            starting_position = \
                starting_grid[entry.driver.index].position
            finish_position = entry.position

            position_lookup[finish_position] = starting_position

        try:
            try:
                font = ImageFont.truetype(
                    kwargs['font'],
                    kwargs['font_size'])
            except (AttributeError, OSError):
                font = ImageFont.load_default()
            font_color = tuple(kwargs['font_color'])
        except KeyError:
            font = ImageFont.load_default()
            font_color = (0, 0, 0)

        self.add_column('position', 'Pos.', colspan=2)
        self._columns.pop(0)
        self._columns.insert(0, self._columns.pop())

        self.add_column(
            'position',
            '',
            formatter=self._make_charm,
            formatter_args={
                'position_lookup': position_lookup,
                'text_height': font.getsize("A")[1],
                'font': font,
                'font_color': font_color})
        self._columns.insert(1, self._columns.pop())

    @staticmethod
    def _position_change(value, **kwargs):
        return "{:+d}".format(kwargs['position_lookup'][value] - value)

    @staticmethod
    def _make_charm(value, **kwargs):
        text_height = kwargs['text_height']
        try:
            change = kwargs['position_lookup'][value] - value
        except KeyError:
            change = 0

        font = kwargs['font']
        charm_width = \
            text_height \
            + font.getsize(str(abs(change)))[0] \
            + 2
        charm_height = text_height + 1

        charm = Image.new(
            'RGBA',
            (charm_width, charm_height),
            (0, 0, 0, 0))

        draw = ImageDraw.Draw(charm)
        if change > 0:
            draw.polygon(
                [
                    (0, text_height),
                    (text_height, text_height),
                    (int(text_height/2), 0)],
                fill=(0, 255, 0, 255),
                outline=(0, 0, 0, 255))
            draw.text(
                (text_height + 2, 0),
                str(abs(change)),
                fill=kwargs['font_color'],
                font=kwargs['font'])
        elif change < 0:
            draw.polygon(
                [
                    (0, 0),
                    (text_height, 0),
                    (int(text_height/2), text_height)],
                fill=(255, 0, 0, 255),
                outline=(0, 0, 0, 255))
            draw.text(
                (text_height + 2, 0),
                str(abs(change)),
                fill=kwargs['font_color'],
                font=kwargs['font'])
        else:
            draw.rectangle(
                [
                    (0, int(text_height*.45)),
                    (text_height, int(text_height*.75))],
                fill=(255, 255, 0, 255),
                outline=(0, 0, 0, 255))
            draw.text(
                (text_height + 2, 0),
                str(abs(change)),
                fill=kwargs['font_color'],
                font=kwargs['font'])

        return charm



"""
Tests REParticipantPacket.py
"""

from hashlib import md5
from struct import pack
import unittest
from unittest.mock import sentinel

from replayenhancer.REParticipantPacket import REParticipantPacket \
    as ParticipantPacket


class TestValidPacket(unittest.TestCase):
    """
    Unit tests for a valid RE Participant Packet.
    """
    expected_build_version = abs(id(sentinel.build_version))
    while expected_build_version > 65535:
        expected_build_version //= 2

    expected_packet_type = 1

    expected_name = list()
    expected_name.extend([str(id(sentinel.name)) for _ in range(16)])

    packet_length = 1347
    packet_string = "HB64x64x64x64x"
    packet_string += "64s" * 16
    packet_string += "64x"

    test_data = list()
    test_data.append(expected_build_version)
    test_data.append(expected_packet_type)
    test_data.extend([name.encode('utf-8') for name in expected_name])

    test_binary_data = pack(packet_string, *test_data)

    def test_init(self):
        instance = ParticipantPacket(self.test_binary_data)
        expected_result = ParticipantPacket
        self.assertIsInstance(instance, expected_result)

    def test_init_wrong_packet_type(self):
        test_data = self.test_data
        test_data[1] = 2  # Set incorrect packet type

        test_binary_data = pack(self.packet_string, *test_data)

        with self.assertRaises(ValueError):
            ParticipantPacket(test_binary_data)

    def test_property_build_version(self):
        instance = ParticipantPacket(self.test_binary_data)
        expected_result = self.expected_build_version
        self.assertEqual(instance.build_version_number, expected_result)

    def test_property_data_hash(self):
        instance = ParticipantPacket(self.test_binary_data)
        expected_result = md5(self.test_binary_data).hexdigest()
        self.assertEqual(instance.data_hash, expected_result)

    def test_property_name(self):
        instance = ParticipantPacket(self.test_binary_data)
        expected_result = type(self.expected_name)
        self.assertIsInstance(instance.name, expected_result)

    def test_property_name_values(self):
        instance = ParticipantPacket(self.test_binary_data)
        expected_result = self.expected_name
        self.assertListEqual(instance.name, expected_result)

    def test_method_repr(self):
        instance = ParticipantPacket(self.test_binary_data)
        expected_result = "REParticipantPacket"
        self.assertEqual(repr(instance), expected_result)

    def test_method_str(self):
        instance = ParticipantPacket(self.test_binary_data)
        expected_result = "REParticipantPacket"
        self.assertEqual(str(instance), expected_result)



"""
Provides a class for the Additional Participant Info Strings output
by Project CARS
"""

from hashlib import md5

from replayenhancer.Packet import Packet


class AdditionalParticipantPacket(Packet):
    """
    Creates an object from an additional participant info string
    packet.

    The additional participant info string packet has a length of
    1028, and is packet type 2.
    """
    def __init__(self, packet_data):
        self.data_hash = md5(packet_data).hexdigest()
        unpacked_data = self._unpack_data(packet_data)

        try:
            self.build_version_number = int(unpacked_data.popleft())

            self._test_packet_type(unpacked_data.popleft())

            self.offset = int(unpacked_data.popleft())

            self.name = list()
            for _ in range(16):
                self.name.append(
                    str(
                        unpacked_data.popleft(),
                        encoding='utf-8',
                        errors='strict').split('\x00', 1)[0])

        except ValueError:
            raise

    @property
    def packet_type(self):
        return 2

    @property
    def _packet_string(self):
        packet_string = "HBB"
        packet_string += "64s"*16

        return packet_string

    def __str__(self):
        return "AdditionalParticipantPacket"



"""
Integration testing of StaticBase.py
"""
import json
import os
import re
import sys

import moviepy.editor as mpy
from moviepy.video.io.bindings import PIL_to_npimage
from PIL import Image, ImageDraw

from replayenhancer.DefaultCards \
    import SeriesStandings, StartingGrid
from replayenhancer.GTStandings import GTStandings
from replayenhancer.RaceData import RaceData
from replayenhancer.RaceResultsWithChange import RaceResultsWithChange
from replayenhancer.SeriesStandingsWithChange \
    import SeriesStandingsWithChange


def make_video(config_file):
    """
    Test of race data.
    """
    configuration = json.load(open(config_file))
    race_data = RaceData(configuration['source_telemetry'])
    result_data = RaceData(configuration['source_telemetry'])
    output_prefix = re.match(
        r'(.*)\.json$',
        os.path.basename(config_file)).group(1)

    if os.environ.get('HEADINGFONTOVERRIDE') is not None:
        configuration['heading_font'] = \
            os.environ['HEADINGFONTOVERRIDE']
    if os.environ.get('DISPLAYFONTOVERRIDE') is not None:
        configuration['font'] = os.environ['DISPLAYFONTOVERRIDE']

    framerate = 30

    source_video = mpy.VideoFileClip(
        configuration['source_video']
    ).subclip(
        configuration['video_skipstart'],
        configuration['video_skipend'])

    pcre_standings = GTStandings(
        race_data,
        ups=framerate,
        **configuration)
    standings_clip_mask = mpy.VideoClip(
        make_frame=pcre_standings.make_mask_frame,
        ismask=True)
    standings_clip = mpy.VideoClip(
        make_frame=pcre_standings.make_frame
    ).set_mask(standings_clip_mask)

    try:
        if sys.argv[2] == "sync":
            def timecode_frame(time):
                """
                Custom make frame for timecode.
                """
                timecode_image = Image.new('RGB', (200, 50))
                draw = ImageDraw.Draw(timecode_image)
                draw.text((50, 25), "%.02f"%(time))
                return PIL_to_npimage(timecode_image)

            timecode_clip = mpy.VideoClip(
                timecode_frame,
                duration=source_video.duration
            ).set_position(('center', 'top'))

            first_lap_data = RaceData(configuration['source_telemetry'])
            while first_lap_data.drivers_by_index[0].laps_complete < 1:
                _ = first_lap_data.get_data()

            main_event = mpy.CompositeVideoClip(
                [source_video, standings_clip, timecode_clip]
            ).set_duration(
                source_video.duration
            ).subclip(
                first_lap_data.elapsed_time-5,
                first_lap_data.elapsed_time+5)
        else:
            raise IndexError

    except IndexError:
        main_event = mpy.CompositeVideoClip(
            [source_video, standings_clip]
        ).set_duration(source_video.duration)

    pcre_starting_grid = StartingGrid(
        sorted(race_data.starting_grid, key=lambda x: x.position),
        size=source_video.size,
        **configuration)
    Image.fromarray(pcre_starting_grid.to_frame()).save(
        output_prefix + "_starting_grid.png")
    starting_grid = mpy.ImageClip(
        pcre_starting_grid.to_frame()).set_duration(5)

    while True:
        try:
            result_data.get_data()
        except StopIteration:
            break

    pcre_results = RaceResultsWithChange(
        sorted(result_data.classification, key=lambda x: x.position),
        result_data.starting_grid,
        size=source_video.size,
        **configuration)
    Image.fromarray(pcre_results.to_frame()).save(
        output_prefix + '_results.png')
    results = mpy.ImageClip(pcre_results.to_frame()).set_duration(20)

    if not any([
            x['points'] for x
            in configuration['participant_config'].values()]):
        pcre_series_standings = SeriesStandings(
            result_data.classification,
            size=source_video.size,
            **configuration)
    else:
        pcre_series_standings = SeriesStandingsWithChange(
            result_data.classification,
            size=source_video.size,
            **configuration)
    Image.fromarray(pcre_series_standings.to_frame()).save(
        output_prefix + '_series_standings.png')
    series_standings = mpy.ImageClip(
        pcre_series_standings.to_frame()).set_duration(20)

    output = mpy.concatenate_videoclips([
        starting_grid.fadeout(1),
        main_event,
        results.fadein(1).fadeout(1),
        series_standings.fadein(1)
    ], method="compose")
    output.write_videofile(configuration['output_video'], fps=framerate)

if __name__ == '__main__':
    make_video(sys.argv[1])



"""
Provides a class for a Series Standings card with position change.
"""
from PIL import Image, ImageDraw, ImageFont

from replayenhancer.DefaultCards import SeriesStandings


class SeriesStandingsWithChange(SeriesStandings):
    """
    Defines a class for a Series Standings card with position change.
    """
    def __init__(self, data, size=None, **kwargs):
        super().__init__(data, size=size, **kwargs)

        try:
            try:
                font = ImageFont.truetype(
                    kwargs['font'],
                    kwargs['font_size'])
            except (AttributeError, OSError):
                font = ImageFont.load_default()
            font_color = tuple(kwargs['font_color'])
        except KeyError:
            font = ImageFont.load_default()
            font_color = (0, 0, 0)

        try:
            point_structure = {
                k: v
                for k, v in enumerate(kwargs['point_structure'])}
        except KeyError:
            point_structure = None

        try:
            points_lookup = {
                k: v['points']
                for k, v in kwargs['participant_config'].items()}
        except KeyError:
            points_lookup = None

        formatter_args = {'point_structure': point_structure,
                          'points_lookup': points_lookup}

        self._columns.pop()
        self.add_column(
            'calc_points_data',
            'Points',
            formatter=self.calc_series_points,
            formatter_args=formatter_args,
            align='center',
            colspan=2)

        self.add_column(
            'calc_points_data',
            '',
            formatter=self._make_charm,
            formatter_args={
                'font': font,
                'font_color': font_color,
                'text_height': font.getsize("A")[1],
                'point_structure': point_structure,
                'points_lookup': points_lookup})

    def _make_charm(self, value, **kwargs):
        #  How many people had more points than us?
        driver_name, _, _ = value
        old_points = kwargs['points_lookup'][driver_name]
        old_more_points = len([
            points for points in kwargs['points_lookup'].values()
            if points > old_points])

        #  How many people have more points than us?
        new_points = int(self.calc_series_points(value, **kwargs))
        current_points = [
            int(self.calc_series_points(
                entry.calc_points_data,
                **kwargs))
            for entry in self._data]
        new_more_points = len([
            points for points in current_points
            if points > new_points])

        change = old_more_points - new_more_points

        text_height = kwargs['text_height']
        charm_width = text_height + 1
        charm_height = text_height + 1

        charm = Image.new(
            'RGBA',
            (charm_width, charm_height),
            (0, 0, 0, 0))

        draw = ImageDraw.Draw(charm)
        if change > 0:
            draw.polygon(
                [
                    (0, text_height),
                    (text_height, text_height),
                    (int(text_height/2), 0)],
                fill=(0, 255, 0, 255),
                outline=(0, 0, 0, 255))
        elif change < 0:
            draw.polygon(
                [
                    (0, 0),
                    (text_height, 0),
                    (int(text_height/2), text_height)],
                fill=(255, 0, 0, 255),
                outline=(0, 0, 0, 255))
        else:
            draw.rectangle(
                [
                    (0, int(text_height*.45)),
                    (text_height, int(text_height*.75))],
                fill=(255, 255, 0, 255),
                outline=(0, 0, 0, 255))

        return charm

    @staticmethod
    def calc_rank(value, **kwargs):
        driver_name, _, _ = value
        ranks = dict()
        last_points = None
        last_rank = 0

        for entry in sorted(
                kwargs['points_lookup'].items(),
                key=lambda x: x[1],
                reverse=True):
            if last_points != entry[1]:
                last_points = entry[1]
                last_rank += 1
            ranks[entry[0]] = last_rank

        return str(ranks[driver_name])



"""Provides a Track class for the Project CARS Replay Enhancer.

Notes
-----
Way too many thanks to mrbelowski, who made much of the data publicly
available through his CrewChief app, preventing me from having to
reinvent the wheel.
https://github.com/mrbelowski/CrewChiefV4/
"""

import os
from json import load


class Track:
    """Represents a Project CARS track.

    Notes
    -----
    Track detection requires the existence of `lib/track_data.json` that
    contains the track data.

    Parameters
    ----------
    track_length : float
        Length of the track, as found in the telemetry data.
    """
    def __init__(self, track_length):
        try:
            with open (os.path.join(
                    os.path.dirname(__file__),
                    'lib/track_data.json')) as json_file:
                json_data = load(json_file)

            matching_tracks = sorted(
                json_data.values(),
                key=lambda x: abs(x['length'] - float(track_length))
            )
            track = matching_tracks[0]

            try:
                # If all the pit keys are present, we populate the pit
                # information.
                self._pit_entry = (
                    float(track['pit_entry'][0]),
                    float(track['pit_entry'][1])
                )
                self._pit_exit = (
                    float(track['pit_exit'][0]),
                    float(track['pit_exit'][1])
                )
                self._pit_radius = float(track['pit_radius'])
            except KeyError:
                self._pit = False
            else:
                self._pit = True
        except FileNotFoundError:
            self._pit = False
        except ValueError as json_error:
            raise json_error

    def at_pit_entry(self, coordinates):
        """Determines if a car is at the pit entry.

        Parameters
        ----------
        coordinates : [float, float, float]
            Car coordinates.

        Returns
        -------
        bool
            True if car is at the pit entry
        """
        return self._pit \
            and abs(self._pit_entry[0]-coordinates[0]) < self._pit_radius \
            and abs(self._pit_entry[1]-coordinates[2]) < self._pit_radius

    def at_pit_exit(self, coordinates):
        """Determines if a car is at the pit exit.

        Parameters
        ----------
        coordinates : [float, float, float]
            Car coordinates.

        Returns
        -------
        bool
            True if car is at the pit exit
        """
        return self._pit \
            and abs(self._pit_exit[0]-coordinates[0]) < self._pit_radius \
            and abs(self._pit_exit[1]-coordinates[2]) < self._pit_radius



"""Tests Track.py

"""

import sys
import unittest
from unittest.mock import mock_open, patch

from replayenhancer.Track import Track


class TestTrack(unittest.TestCase):
    """Unit tests against the Track object.

    """
    track_data = {
        "Test Track:Short": {
            "display_name": "Test Track Short",
            "length": 42.0,
            "pit_entry": [0, 0],
            "pit_exit": [100, 100],
            "pit_radius": 2
        },

        "Test Track:Kart": {
            "display_name": "Test Track Kart",
            "length": 15.0
        }
    }

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_init(self):
        with patch('replayenhancer.Track.open', mock_open()), \
                patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            expected_result = Track
            self.assertIsInstance(instance, expected_result)

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_init_pre35(self):
        with patch('builtins.open', mock_open()), \
                patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            expected_result = Track
            self.assertIsInstance(instance, expected_result)

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_init_no_json(self):
        with patch('replayenhancer.Track.open', mock_open()) as mock_open_error:
            mock_open_error.side_effect = FileNotFoundError
            instance = Track(42.0)
            expected_result = Track
            self.assertIsInstance(instance, expected_result)

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_init_no_json_pre35(self):
        with patch('builtins.open', mock_open()) as mock_open_error:
            mock_open_error.side_effect = FileNotFoundError
            instance = Track(42.0)
            expected_result = Track
            self.assertIsInstance(instance, expected_result)

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_init_bad_json(self):
        with patch('replayenhancer.Track.open', mock_open()), \
                patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.side_effect = ValueError
            with self.assertRaises(ValueError):
                _ = Track(42.0)

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_init_bad_json_pre35(self):
        with patch('builtins.open', mock_open()), \
                patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.side_effect = ValueError
            with self.assertRaises(ValueError):
                _ = Track(42.0)

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_at_pit_entry_exact_true(self):
        with patch('replayenhancer.Track.open', mock_open()), \
                patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertTrue(instance.at_pit_entry([0, -99, 0]))

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_at_pit_entry_exact_true_pre35(self):
        with patch('builtins.open', mock_open()), \
                patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertTrue(instance.at_pit_entry([0, -99, 0]))

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_at_pit_entry_close_x_true(self):
        with patch('replayenhancer.Track.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertTrue(instance.at_pit_entry([1, -99, 0]))

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_at_pit_entry_close_x_true_pre35(self):
        with patch('builtins.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertTrue(instance.at_pit_entry([1, -99, 0]))

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_at_pit_entry_edge_x_false(self):
        with patch('replayenhancer.Track.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertFalse(instance.at_pit_entry([2, -99, 0]))

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_at_pit_entry_edge_x_false_pre35(self):
        with patch('builtins.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertFalse(instance.at_pit_entry([2, -99, 0]))

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_at_pit_entry_far_x_false(self):
        with patch('replayenhancer.Track.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertFalse(instance.at_pit_entry([3, -99, 0]))

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_at_pit_entry_far_x_false_pre35(self):
        with patch('builtins.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertFalse(instance.at_pit_entry([3, -99, 0]))

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_at_pit_entry_close_z_true(self):
        with patch('replayenhancer.Track.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertTrue(instance.at_pit_entry([0, -99, 1]))

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_at_pit_entry_close_z_true_pre35(self):
        with patch('builtins.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertTrue(instance.at_pit_entry([0, -99, 1]))

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_at_pit_entry_exact_z_false(self):
        with patch('replayenhancer.Track.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertFalse(instance.at_pit_entry([0, -99, 2]))

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_at_pit_entry_exact_z_false_pre35(self):
        with patch('builtins.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertFalse(instance.at_pit_entry([0, -99, 2]))

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_at_pit_entry_far_z_false(self):
        with patch('replayenhancer.Track.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertFalse(instance.at_pit_entry([0, -99, 3]))

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_at_pit_entry_far_z_false_pre35(self):
        with patch('builtins.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertFalse(instance.at_pit_entry([0, -99, 3]))

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_at_pit_entry_no_pits(self):
        with patch('replayenhancer.Track.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(15.0)
            self.assertFalse(instance.at_pit_entry([0, -99, 0]))

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_at_pit_entry_no_pits_pre35(self):
        with patch('builtins.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(15.0)
            self.assertFalse(instance.at_pit_entry([0, -99, 0]))

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_at_pit_entry_no_json(self):
        with patch('replayenhancer.Track.open', mock_open()) as mock_open_error:
            mock_open_error.side_effect = FileNotFoundError
            instance = Track(42.0)
            self.assertFalse(instance.at_pit_entry([0, -99, 0]))

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_at_pit_entry_no_json_pre35(self):
        with patch('builtins.open', mock_open()) as mock_open_error:
            mock_open_error.side_effect = FileNotFoundError
            instance = Track(42.0)
            self.assertFalse(instance.at_pit_entry([0, -99, 0]))

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_at_pit_exit_exact_true(self):
        with patch('replayenhancer.Track.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertTrue(instance.at_pit_exit([100, -99, 100]))

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_at_pit_exit_exact_true_pre35(self):
        with patch('builtins.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertTrue(instance.at_pit_exit([100, -99, 100]))

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_at_pit_exit_close_x_true(self):
        with patch('replayenhancer.Track.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertTrue(instance.at_pit_exit([101, -99, 100]))

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_at_pit_exit_close_x_true_pre35(self):
        with patch('builtins.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertTrue(instance.at_pit_exit([101, -99, 100]))

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_at_pit_exit_edge_x_false(self):
        with patch('replayenhancer.Track.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertFalse(instance.at_pit_exit([102, -99, 100]))

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_at_pit_exit_edge_x_false_pre35(self):
        with patch('builtins.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertFalse(instance.at_pit_exit([102, -99, 100]))

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_at_pit_exit_far_x_false(self):
        with patch('replayenhancer.Track.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertFalse(instance.at_pit_exit([103, -99, 100]))

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_at_pit_exit_far_x_false_pre35(self):
        with patch('builtins.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertFalse(instance.at_pit_exit([103, -99, 100]))

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_at_pit_exit_close_z_true(self):
        with patch('replayenhancer.Track.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertTrue(instance.at_pit_exit([100, -99, 101]))

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_at_pit_exit_close_z_true_pre35(self):
        with patch('builtins.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertTrue(instance.at_pit_exit([100, -99, 101]))

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_at_pit_exit_edge_z_false(self):
        with patch('replayenhancer.Track.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertFalse(instance.at_pit_exit([100, -99, 102]))

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_at_pit_exit_edge_z_false_pre35(self):
        with patch('builtins.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertFalse(instance.at_pit_exit([100, -99, 102]))

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_at_pit_exit_far_z_false(self):
        with patch('replayenhancer.Track.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertFalse(instance.at_pit_exit([100, -99, 103]))

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_at_pit_exit_far_z_false_pre35(self):
        with patch('builtins.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(42.0)
            self.assertFalse(instance.at_pit_exit([100, -99, 103]))

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_at_pit_exit_no_pits(self):
        with patch('replayenhancer.Track.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(15.0)
            self.assertFalse(instance.at_pit_exit([100, -99, 100]))

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_at_pit_exit_no_pits_pre35(self):
        with patch('builtins.open', mock_open()), \
             patch('replayenhancer.Track.load') as mock_json_load:
            mock_json_load.return_value = self.track_data
            instance = Track(15.0)
            self.assertFalse(instance.at_pit_exit([100, -99, 100]))

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_at_pit_exit_no_json(self):
        with patch('replayenhancer.Track.open', mock_open()) as mock_open_error:
            mock_open_error.side_effect = FileNotFoundError
            instance = Track(42.0)
            self.assertFalse(instance.at_pit_exit([0, -99, 0]))

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_at_pit_exit_no_json_pre35(self):
        with patch('builtins.open', mock_open()) as mock_open_error:
            mock_open_error.side_effect = FileNotFoundError
            instance = Track(42.0)
            self.assertFalse(instance.at_pit_exit([0, -99, 0]))



"""
Tests DefaultCards.py
"""

import unittest
from unittest.mock import MagicMock, PropertyMock, patch, sentinel

import numpy
from PIL import ImageFont

from replayenhancer.DefaultCards \
    import RaceResults, SeriesStandings, StartingGrid, SeriesChampion


class TestRaceResults(unittest.TestCase):
    """
    Unit tests for Race Results card.
    """

    @patch('replayenhancer.RaceData.Driver', autospec=True)
    @patch('replayenhancer.RaceData.ClassificationEntry', autospec=True)
    def test_init_no_config(self, mock_classification_entry, mock_driver):
        mock_driver.laps_complete = 6
        mock_driver.race_time = 42.00
        mock_driver.stops = 0
        mock_classification_entry.driver = mock_driver
        instance = RaceResults([mock_classification_entry])
        expected_result = RaceResults
        self.assertIsInstance(instance, expected_result)

    @patch('replayenhancer.RaceData.Driver', autospec=True)
    @patch('replayenhancer.RaceData.ClassificationEntry', autospec=True)
    def test_init_config(self, mock_classification_entry, mock_driver):
        configuration = {
            'participant_config': {
                'Kobernulf Monnur': {
                    'display': 'Senor Pez',
                    'car': '125cc Shifter Kart',
                    'team': 'DarkNitro',
                }
            },
            'point_structure': [5, 15, 12, 10, 8, 6, 4, 2, 1]
        }
        mock_driver.laps_complete = 6
        mock_driver.race_time = 42.00
        mock_driver.stops = 0
        mock_classification_entry.driver = mock_driver
        instance = RaceResults([mock_classification_entry], **configuration)
        expected_result = RaceResults
        self.assertIsInstance(instance, expected_result)

    @patch('replayenhancer.RaceData.Driver', autospec=True)
    @patch('replayenhancer.RaceData.ClassificationEntry', autospec=True)
    def test_method_calc_points_best_lap(self, mock_classification_entry,
                                         mock_driver):
        driver_name = 'Kobernulf_Monnur'
        position = 1
        best_lap = 42.0

        configuration = {
            'point_structure': [5, 15, 12, 10, 8, 6, 4, 2, 1]
        }
        type(mock_classification_entry).best_lap = PropertyMock(
            return_value=best_lap)

        mock_driver.laps_complete = 6
        mock_driver.race_time = 42.00
        mock_driver.stops = 0
        mock_classification_entry.driver = mock_driver

        instance = RaceResults([mock_classification_entry], **configuration)
        expected_result = '20'
        self.assertEqual(
            instance.calc_points(
                (driver_name, position, best_lap),
                **configuration),
            expected_result)

    @patch('replayenhancer.RaceData.Driver', autospec=True)
    @patch('replayenhancer.RaceData.ClassificationEntry', autospec=True)
    def test_method_calc_points_not_best_lap(self, mock_classification_entry,
                                             mock_driver):
        driver_name = 'Kobernulf Monnur'
        position = 1
        best_lap = 56.0

        configuration = {
            'point_structure': [5, 15, 12, 10, 8, 6, 4, 2, 1]
        }
        type(mock_classification_entry).best_lap = PropertyMock(
            return_value=42.0)

        mock_driver.laps_complete = 6
        mock_driver.race_time = 42.00
        mock_driver.stops = 0
        mock_classification_entry.driver = mock_driver

        instance = RaceResults([mock_classification_entry], **configuration)
        expected_result = '15'
        self.assertEqual(
            instance.calc_points(
                (driver_name, position, best_lap),
                **configuration),
            expected_result)

    @patch('replayenhancer.RaceData.Driver', autospec=True)
    @patch('replayenhancer.RaceData.ClassificationEntry', autospec=True)
    def test_method_calc_points_no_point_structure(
            self, mock_classification_entry, mock_driver):
        driver_name = 'Kobernulf Monnur'
        position = 1
        best_lap = 42.0

        mock_driver.laps_complete = 6
        mock_driver.race_time = 42.00
        mock_driver.stops = 0
        mock_classification_entry.driver = mock_driver
        instance = RaceResults([mock_classification_entry])
        expected_result = '0'
        self.assertEqual(
            instance.calc_points((driver_name, position, best_lap)),
            expected_result)

    def test_method_format_time_below_min(self):
        time = 42
        expected_result = '42.000'
        self.assertEqual(RaceResults.format_time(time), expected_result)

    def test_method_format_time_below_min_truncate(self):
        time = 42.1234
        expected_result = '42.123'
        self.assertEqual(RaceResults.format_time(time), expected_result)

    def test_method_format_time_below_min_round(self):
        time = 42.9876
        expected_result = '42.988'
        self.assertEqual(RaceResults.format_time(time), expected_result)

    def test_method_format_time_below_hour(self):
        time = 84
        expected_result = '1:24.000'
        self.assertEqual(RaceResults.format_time(time), expected_result)

    def test_method_format_time_below_hour_truncate(self):
        time = 84.1234
        expected_result = '1:24.123'
        self.assertEqual(RaceResults.format_time(time), expected_result)

    def test_method_format_time_below_hour_round(self):
        time = 84.9876
        expected_result = '1:24.988'
        self.assertEqual(RaceResults.format_time(time), expected_result)

    def test_method_format_time(self):
        time = 3702
        expected_result = '1:01:42.000'
        self.assertEqual(RaceResults.format_time(time), expected_result)

    def test_method_format_time_truncate(self):
        time = 3702.1234
        expected_result = '1:01:42.123'
        self.assertEqual(RaceResults.format_time(time), expected_result)

    def test_method_format_time_round(self):
        time = 3702.9876
        expected_result = '1:01:42.988'
        self.assertEqual(RaceResults.format_time(time), expected_result)

    def test_method_format_time_passed_none(self):
        time = None
        expected_result = ""
        self.assertEqual(RaceResults.format_time(time), expected_result)

    def test_method_format_time_passed_string(self):
        time = "ERROR"
        expected_result = ""
        self.assertEqual(RaceResults.format_time(time), expected_result)


class TestStartingGrid(unittest.TestCase):
    """
    Unit tests for Starting Grid card.
    """

    @patch('replayenhancer.RaceData.StartingGridEntry', autospec=True)
    def test_init_no_config(self, mock_starting_grid_entry):
        instance = StartingGrid(mock_starting_grid_entry)
        expected_result = StartingGrid
        self.assertIsInstance(instance, expected_result)

    @patch('replayenhancer.RaceData.StartingGridEntry', autospec=True)
    def test_init_config(self, mock_starting_grid_entry):
        configuration = {
            'participant_config': {
                'Kobernulf Monnur': {
                    'display': 'Senor Pez',
                    'car': '125cc Shifter Kart',
                    'team': 'DarkNitro',
                    'points': 0
                }
            }
        }
        instance = StartingGrid([mock_starting_grid_entry], **configuration)
        expected_result = StartingGrid
        self.assertIsInstance(instance, expected_result)


class TestSeriesStandings(unittest.TestCase):
    """
    Unit Tests for Series Standings card.
    """

    @patch('replayenhancer.StaticBase.StaticBase.sort_data', autospec=True)
    @patch('replayenhancer.RaceData.ClassificationEntry', autospec=True)
    def test_init_no_config(self, mock_classification_entry, mock_sort_data):
        mock_sort_data.return_value = ('Kobernulf Monnur', 1, 42.0)

        instance = SeriesStandings(mock_classification_entry)
        expected_result = SeriesStandings
        self.assertIsInstance(instance, expected_result)

    @patch('replayenhancer.StaticBase.StaticBase.sort_data', autospec=True)
    @patch('replayenhancer.RaceData.ClassificationEntry', autospec=True)
    def test_init_config(self, mock_classification_entry, mock_sort_data):
        mock_sort_data.return_value = ('Kobernulf Monnur', 1, 42.0)

        configuration = {
            'participant_config': {
                'Kobernulf Monnur': {
                    'display': 'Senor Pez',
                    'car': '125cc Shifter Kart',
                    'team': 'Dark Nitro',
                    'points': 15
                }
            },
            'point_structure': [5, 15, 12, 10, 8, 6, 4, 2, 1]
        }

        instance = SeriesStandings([mock_classification_entry], **configuration)
        expected_result = SeriesStandings
        self.assertIsInstance(instance, expected_result)

    @patch('replayenhancer.StaticBase.StaticBase.sort_data', autospec=True)
    @patch('replayenhancer.RaceData.ClassificationEntry', autospec=True)
    def test_method_calc_series_points(self, mock_classification_entry, mock_sort_data):
        driver_name = 'Kobernulf Monnur'
        position = 1
        best_lap = 42.0

        mock_sort_data.return_value = (driver_name, position, best_lap)
        type(mock_classification_entry).best_lap = PropertyMock(
            return_value=best_lap)

        configuration = {
            'points_lookup': {
                'Kobernulf Monnur': 15
            },
            'point_structure': [5, 15, 12, 10, 8, 6, 4, 2, 1]
        }

        instance = SeriesStandings([mock_classification_entry])
        expected_result = '35'
        self.assertEqual(
            instance.calc_series_points(
                (driver_name, position, best_lap),
                **configuration),
            expected_result)

    @patch('replayenhancer.StaticBase.StaticBase.sort_data', autospec=True)
    @patch('replayenhancer.RaceData.ClassificationEntry', autospec=True)
    def test_method_calc_series_points_no_entry(self, mock_classification_entry, mock_sort_data):
        driver_name = 'Kobernulf Monnur'
        position = 1
        best_lap = 42.0

        mock_sort_data.return_value = (driver_name, position, best_lap)
        type(mock_classification_entry).best_lap = PropertyMock(
            return_value=best_lap)

        configuration = {
            'point_structure': [5, 15, 12, 10, 8, 6, 4, 2, 1]
        }

        instance = SeriesStandings([mock_classification_entry])
        expected_result = '20'
        self.assertEqual(
            instance.calc_series_points(
                (driver_name, position, best_lap),
                **configuration),
            expected_result)

    @patch('replayenhancer.StaticBase.StaticBase.sort_data', autospec=True)
    @patch('replayenhancer.RaceData.ClassificationEntry', autospec=True)
    def test_method_calc_series_rank_first(self, mock_classification_entry, mock_sort_data):
        driver_name = 'Kobernulf Monnur'
        position = 1
        best_lap = 42.0

        mock_sort_data.return_value = (driver_name, position, best_lap)
        type(mock_classification_entry).best_lap = PropertyMock(
            return_value=best_lap)
        type(mock_classification_entry).calc_points_data = PropertyMock(
            return_value=(driver_name, position, best_lap))
        type(mock_classification_entry).driver_name = PropertyMock(
            return_value=driver_name)

        configuration = {
            'points_lookup': {
                'Kobernulf Monnur': 15
            },
            'point_structure': [5, 15, 12, 10, 8, 6, 4, 2, 1]
        }

        instance = SeriesStandings([mock_classification_entry])
        expected_result = '1'
        self.assertEqual(
            instance.calc_series_rank(
                (driver_name, position, best_lap),
                **configuration),
            expected_result)

    @patch('replayenhancer.StaticBase.StaticBase.sort_data', autospec=True)
    @patch('replayenhancer.RaceData.ClassificationEntry')
    def test_method_calc_series_rank_tie_second(self, mock_classification_entry, mock_sort_data):
        mock_sort_data.return_value = [
            ('First Place', 1, 42.0),
            ('Second Place', 2, 42.0),
            ('Third Place', 3, 42.0)
        ]

        patcher = patch(
            'replayenhancer.RaceData.ClassificationEntry',
            best_lap=42.0,
            calc_points_data=('First Place', 1, 42.0),
            driver_name='First Place')
        first_place = patcher.start()

        patcher = patch(
            'replayenhancer.RaceData.ClassificationEntry',
            best_lap=42.0,
            calc_points_data=('Second Place', 2, 42.0),
            driver_name='Second Place')
        second_place = patcher.start()

        patcher = patch(
            'replayenhancer.RaceData.ClassificationEntry',
            best_lap=42.0,
            calc_points_data=('Third Place', 3, 42.0),
            driver_name='Third Place')
        third_place = patcher.start()

        configuration = {
            'points_lookup': {
                'First Place': 10,
                'Second Place': 5,
                'Third Place': 5
            },
            'point_structure': [0, 10, 6, 6]
        }

        instance = SeriesStandings([first_place, second_place, third_place])
        expected_result = '2'
        self.assertEqual(
            instance.calc_series_rank(
                ('Third Place', 3, 42.0),
                **configuration),
            expected_result)\

    @patch('replayenhancer.StaticBase.StaticBase.sort_data', autospec=True)
    @patch('replayenhancer.RaceData.ClassificationEntry')
    def test_method_calc_series_rank_third(self, mock_classification_entry, mock_sort_data):
        mock_sort_data.return_value = [
            ('First Place', 1, 42.0),
            ('Second Place', 2, 42.0),
            ('Third Place', 3, 42.0)
        ]

        patcher = patch(
            'replayenhancer.RaceData.ClassificationEntry',
            best_lap=42.0,
            calc_points_data=('First Place', 1, 42.0),
            driver_name='First Place')
        first_place = patcher.start()

        patcher = patch(
            'replayenhancer.RaceData.ClassificationEntry',
            best_lap=42.0,
            calc_points_data=('Second Place', 2, 42.0),
            driver_name='Second Place')
        second_place = patcher.start()

        patcher = patch(
            'replayenhancer.RaceData.ClassificationEntry',
            best_lap=42.0,
            calc_points_data=('Third Place', 3, 42.0),
            driver_name='Third Place')
        third_place = patcher.start()

        configuration = {
            'points_lookup': {
                'First Place': 10,
                'Second Place': 10,
                'Third Place': 5
            },
            'point_structure': [0, 10, 10, 6]
        }

        instance = SeriesStandings([first_place, second_place, third_place])
        expected_result = '3'
        self.assertEqual(
            instance.calc_series_rank(
                ('Third Place', 3, 42.0),
                **configuration),
            expected_result)


class TestSeriesChampion(unittest.TestCase):
    """
    Unit tests for Series Champion card.
    """

    @patch('replayenhancer.StaticBase.StaticBase.sort_data', autospec=True)
    @patch('replayenhancer.RaceData.ClassificationEntry', autospec=True)
    def test_init_no_config(self, mock_classification_entry, mock_sort_data):
        mock_sort_data.return_value = ('Kobernulf Monnur', 1, 42.0)

        instance = SeriesChampion([mock_classification_entry])
        expected_result = SeriesChampion
        self.assertIsInstance(instance, expected_result)

    @patch('replayenhancer.StaticBase.StaticBase.sort_data', autospec=True)
    @patch('replayenhancer.RaceData.ClassificationEntry', autospec=True)
    def test_init_config(self, mock_classification_entry, mock_sort_data):
        mock_sort_data.return_value = ('Kobernulf Monnur', 1,  42.0)

        configuration = {
            'participant_config': {
                'Kobernulf Monnur': {
                    'display': 'Senor Pez',
                    'car': '125cc Shifter Kart',
                    'team': 'Dark Nitro',
                    'points': 15
                }
            },
            'point_structure': [5, 15, 12, 10, 8, 6, 4, 2, 1]
        }

        instance = SeriesChampion([mock_classification_entry], **configuration)
        expected_result = SeriesChampion
        self.assertIsInstance(instance, expected_result)

    @patch('replayenhancer.StaticBase.StaticBase.sort_data', autospec=True)
    @patch('replayenhancer.RaceData.ClassificationEntry', autospec=True)
    def test_method_to_frame_no_header(self, mock_classification_entry, mock_sort_data):
        first_place = mock_classification_entry
        type(first_place).calc_points_data = PropertyMock(return_value=(
            'Kobernulf Monnur',
            1,
            42.0))
        type(first_place).driver_name = PropertyMock(return_value=(
            'Kobernulf Monnur'))

        second_place = mock_classification_entry
        type(second_place).calc_points_data = PropertyMock(
            return_value=(
                'Second Place',
                2,
                43.0))
        type(second_place).driver_name = PropertyMock(return_value=(
            'Second Place'))

        third_place = mock_classification_entry
        type(third_place).calc_points_data = PropertyMock(return_value=(
            'Third Place',
            3,
            43.5))
        type(third_place).driver_name = PropertyMock(return_value=(
            'Third Place'))

        mock_sort_data.return_value = [
            ('Kobernulf Monnur', 1, 42.0),
            ('Second Place', 2, 43.0),
            ('Third Place', 3, 43.5)]

        expected_value = numpy.ndarray
        instance = SeriesChampion([first_place, second_place, third_place])
        self.assertIsInstance(instance.to_frame(), expected_value)

    @patch('replayenhancer.StaticBase.StaticBase.sort_data', autospec=True)
    @patch('replayenhancer.RaceData.ClassificationEntry', autospec=True)
    def test_method_to_frame_blank_header(self, mock_classification_entry, mock_sort_data):
        first_place = mock_classification_entry
        type(first_place).calc_points_data = PropertyMock(return_value=(
            'Kobernulf Monnur',
            1,
            42.0))
        type(first_place).driver_name = PropertyMock(return_value=(
            'Kobernulf Monnur'))

        second_place = mock_classification_entry
        type(second_place).calc_points_data = PropertyMock(
            return_value=(
                'Second Place',
                2,
                43.0))
        type(second_place).driver_name = PropertyMock(return_value=(
            'Second Place'))

        third_place = mock_classification_entry
        type(third_place).calc_points_data = PropertyMock(return_value=(
            'Third Place',
            3,
            43.5))
        type(third_place).driver_name = PropertyMock(return_value=(
            'Third Place'))

        mock_sort_data.return_value = [
            ('Kobernulf Monnur', 1, 42.0),
            ('Second Place', 2, 43.0),
            ('Third Place', 3, 43.5)]

        configuration = {
            'heading_color': (255, 0, 0)}

        expected_value = numpy.ndarray
        instance = SeriesChampion([first_place, second_place, third_place], **configuration)
        self.assertIsInstance(instance.to_frame(), expected_value)


import csv
import curses
import sys
import time

def displayPosition(stdscr, tData, pData):
	startTime = -1

	for data in tData:
		stdscr.redrawwin()
		#Race not yet started.
		if int(data[9]) & int('111', 2) == 1:
			#stdscr.clear()
			stdscr.addstr(0, 0, "Waiting...")
			stdscr.refresh()
		#Race underway
		elif int(data[9]) & int('111', 2) == 2:
			if startTime == -1:
				startTime = time.time()
				stdscr.clear()
				stdscr.addstr(0, 0, "Press any key to begin race!")
				stdscr.refresh()
				stdscr.getkey()
				stdscr.clear()

				sectorBests = [-1, -1, -1]
				personalBests = [[-1, -1, -1] for x in range(64)]
				bestLap = -1
				personalBestLaps = [-1 for x in range(64)]
				lastLaps = [[-1, -1, -1] for x in range(64)]
				curses.init_pair(1, curses.COLOR_MAGENTA, curses.COLOR_BLACK)
				curses.init_pair(2, curses.COLOR_GREEN, curses.COLOR_BLACK)

			standings = sorted({(int(data[182+i*9]), n, int(data[184+i*9]), i) for (i, n) in pData})
			#stdscr.clear()
			stdscr.addstr(0, 0, "Current Lap: {}".format(max(standings, key=lambda x: x[2])[2]))
			stdscr.addstr(1, 0, "Current Time: {:.3f}".format(float(data[13])))
			for i, standing in enumerate(standings, 2):
				stdscr.addstr(i, 0, "{:>2}: {:<30.30s}".format(int(standing[0]) & int('01111111', 2), standing[1]))
				sector = int(data[185+standing[3]*9]) & int('111', 2)
				stdscr.addstr(i, 15+sector*20, "{:20.20s}".format(' '))
				
				if float(data[186+standing[3]*9]) != -123:
					if sector == 1:
						printSector = 3
						if lastLaps[standing[3]][0] != float(data[186+standing[3]*9]):
							lastLaps[standing[3]][0] = float(data[186+standing[3]*9])

						if bestLap == -1 or bestLap >= sum(lastLaps[standing[3]]):
							bestLap = sum(lastLaps[standing[3]])
							personalBestLaps[standing[3]] = sum(lastLaps[standing[3]])
							stdscr.addstr(i, 15+20*4, "{:^20.3f}".format(float(sum(lastLaps[standing[3]]))), curses.color_pair(1))
						elif personalBestLaps[standing[3]] == -1 or personalBestLaps[standing[3]] >= sum(lastLaps[standing[3]]):
							personalBestLaps[standing[3]] = sum(lastLaps[standing[3]])
							stdscr.addstr(i, 15+20*4, "{:^20.3f}".format(float(sum(lastLaps[standing[3]]))), curses.color_pair(2))
						else:
							stdscr.addstr(i, 15+20*4, "{:^20.3f}".format(float(sum(lastLaps[standing[3]]))))
					else:
						printSector = sector-1

						if sector == 2:
							if lastLaps[standing[3]][1] != float(data[186+standing[3]*9]):
								lastLaps[standing[3]][1] = float(data[186+standing[3]*9])
						else:
							if lastLaps[standing[3]][2] != float(data[186+standing[3]*9]):
								lastLaps[standing[3]][2] = float(data[186+standing[3]*9])

					if sectorBests[printSector-1] == -1 or sectorBests[printSector-1] >= float(data[186+standing[3]*9]):
						sectorBests[printSector-1] = float(data[186+standing[3]*9])
						personalBests[standing[3]][printSector-1] = float(data[186+standing[3]*9])
						stdscr.addstr(i, 15+20*printSector, "{:^20.3f}".format(float(data[186+standing[3]*9])), curses.color_pair(1))
					elif personalBests[standing[3]][printSector-1] == -1 or personalBests[standing[3]][printSector-1] >= float(data[186+standing[3]*9]):
						personalBests[standing[3]][printSector-1] = float(data[186+standing[3]*9])
						stdscr.addstr(i, 15+20*printSector, "{:^20.3f}".format(float(data[186+standing[3]*9])), curses.color_pair(2))
					else:
						stdscr.addstr(i, 15+20*printSector, "{:^20.3f}".format(float(data[186+standing[3]*9])))

			stdscr.addstr(16, 0, "{:>32}".format("Sector Bests:"))
			for n, t in [(x, y) for x, y in enumerate(sectorBests, 1) if y != -1]:
				stdscr.addstr(16, 15+20*n, "{:^20.3f}".format(float(t)))

			if bestLap != -1:
				stdscr.addstr(16, 15+20*4, "{:^20.3f}".format(float(bestLap)))
			
			stdscr.refresh()

	#All done.
	standings = sorted({(int(data[182+i*9]), n, int(data[184+i*9]), i) for (i, n) in pData})
	if sectorBests[2] == -1 or sectorBests[2] >= float(data[186+standings[0][3]*9]):
		sectorBests[2] = float(data[186+standings[0][3]*9])
		personalBests[standings[0][3]][2] = float(data[186+standings[0][3]*9])
		stdscr.addstr(2, 15+20*3, "{:^20.3f}".format(float(data[186+standings[0][3]*9])), curses.color_pair(1))
	elif personalBests[standings[0][3]][2] == -1 or personalBests[standings[0][3]][2] >= float(data[186+standings[0][3]*9]):
		personalBests[standings[0][3]][2] = float(data[186+standings[0][3]*9])
		stdscr.addstr(2, 15+20*3, "{:^20.3f}".format(float(data[186+standings[0][3]*9])), curses.color_pair(2))
	else:
		stdscr.addstr(2, 15+20*3, "{:^20.3f}".format(float(data[186+standings[0][3]*9])))
	
	stdscr.move(1, 0)
	stdscr.clrtoeol()
	stdscr.addstr(1, 0, "Race finished. Press any key to exit!")
	stdscr.refresh()
	stdscr.getkey()
	

if len(sys.argv) == 1:
	print ("Usage: 'python "+sys.argv[0]+" <packetdirectory>'")

else:
	try:
		with open('./'+sys.argv[1]+'tele.csv', 'rb') as f:
			csvdata = csv.reader(f)

			#Split packets into type 0 (telemetry) and types 1 & 2 (participants)
			#tData, pData, aData = zip(*[itertools.izip_longest(t, p, a) for t in csvdata if int(t[1]) & 3 == 0 for x in csvdata if int(x[1]) & 3 == 1 for p in enumerate(x[6:]) if len(p[1]) for y in csvdata if int(y[1]) & 3 == 2 for a in enumerate(y[3:], y[2]) if len(a[1])])

			tData = list()
			pData = list()
			
			for row in csvdata:
				if int(row[1]) & 3 == 0:
					tData.append(row)
				elif int(row[1]) & 3 == 1:
					for p in enumerate(row[6:]):
						if len(p[1]): 
							pData.append(p)
				elif int(row[1]) & 3 == 2:
					for p in enumerate(row[3:], row[2]):
						if len(p[1]):
							pData.append(p)
				else:
					raise ValueError("ValueError: Unrecognized packet type ("+str(int(row[1]) & 3)+")")

		#Because I'm a dirty savescummer, we need to findc the last race, since
		#the earlier ones are savescum scum. Save.
		raceStart = -1
		raceEnd = -1
		for i, data in reversed(list(enumerate(tData))):
			if (int(data[9]) & int('111', 2) == 3) and (raceEnd == -1):
				raceEnd = i+1

			if (int(data[9]) & int('111', 2) == 0) and (raceEnd != -1) and (raceStart == -1):
				raceStart = i+1

		if raceStart == -1 or raceEnd == -1:
			#raise ValueError("ValueError: Couldn't detect raceStart and raceEnd you savescumming scum.")
			raceStart = 0
			raceEnd = -1

		tData = tData[raceStart:raceEnd]
		curses.wrapper(displayPosition, tData, pData)
	
	except ValueError as e:
		print >> sys.stderr, e.message

	except IOError as e:
		print >> sys.stderr, "IOError: %s" % e



"""Tests ParticipantPacket.py

"""
import unittest
from hashlib import md5
from struct import pack

from replayenhancer.ParticipantPacket import ParticipantPacket


class TestParticipantPacket(unittest.TestCase):
    """Unit tests against the ParticipantPacket object.

    """
    expected_build_version_number = 12345
    expected_packet_type = 1

    expected_car_name = "F1 W07 Hybrid"
    expected_car_class_name = "F1 2016"
    expected_track_location = "Abu Dhabi"
    expected_track_variation = "Grand Prix"

    expected_name = [
        "Nico Rosberg",
        "Lewis Hamilton",
        "Daniel Ricciardo",
        "Sebastian Vettel",
        "Max Verstappen",
        "Kimi Räikkönen",
        "Sergio Pérez",
        "Valtteri Bottas",
        "Nico Hülkenberg",
        "Fernando Alonso",
        "Felipe Massa",
        "Carlos Sainz Jr.",
        "Romain Grosjean",
        "Daniil Kvyat",
        "Jenson Button",
        "Kevin Magnussen"
    ]

    expected_packet_length = 1347

    @classmethod
    def binary_data(cls, **kwargs):
        test_data = list()
        packet_string = "HB64s64s64s64s"
        packet_string += "64s" * 16
        packet_string += "64x"

        try:
            test_data.append(kwargs['build_version_number'])
        except KeyError:
            test_data.append(cls.expected_build_version_number)

        try:
            test_data.append(kwargs['packet_type'])
        except KeyError:
            test_data.append(cls.expected_packet_type)

        try:
            test_data.append(kwargs['car_name'].encode('utf-8'))
        except KeyError:
            test_data.append(cls.expected_car_name.encode('utf-8'))

        try:
            test_data.append(kwargs['car_class_name'].encode('utf-8'))
        except KeyError:
            test_data.append(cls.expected_car_class_name.encode('utf-8'))

        try:
            test_data.append(kwargs['track_location'].encode('utf-8'))
        except KeyError:
            test_data.append(cls.expected_track_location.encode('utf-8'))

        try:
            test_data.append(kwargs['track_variation'].encode('utf-8'))
        except KeyError:
            test_data.append(cls.expected_track_variation.encode('utf-8'))

        try:
            test_data.extend([name.encode('utf-8') for name in kwargs['name']])
        except KeyError:
            test_data.extend(
                [name.encode('utf-8') for name in cls.expected_name])

        return pack(packet_string, *test_data)

    def test_init(self):
        instance = ParticipantPacket(self.binary_data())
        expected_result = ParticipantPacket
        self.assertIsInstance(instance, expected_result)

    def test_init_wrong_packet_length(self):
        test_binary_data = pack("H", 42)

        from struct import error
        with self.assertRaises(error):
            ParticipantPacket(test_binary_data)

    def test_init_wrong_packet_type(self):
        with self.assertRaises(ValueError):
            ParticipantPacket(self.binary_data(packet_type=2))

    def test_property_build_version_number(self):
        instance = ParticipantPacket(self.binary_data())
        expected_result = self.expected_build_version_number
        self.assertEqual(instance.build_version_number, expected_result)

    def test_property_car_class_name(self):
        instance = ParticipantPacket(self.binary_data())
        expected_result = self.expected_car_class_name
        self.assertEqual(instance.car_class_name, expected_result)

    def test_property_car_class_name_split_on_null(self):
        instance = ParticipantPacket(self.binary_data(
            car_class_name=self.expected_car_class_name + "\x00Garbage Data"))
        expected_result = self.expected_car_class_name
        self.assertEqual(instance.car_class_name, expected_result)

    def test_property_car_name(self):
        instance = ParticipantPacket(self.binary_data())
        expected_result = self.expected_car_name
        self.assertEqual(instance.car_name, expected_result)

    def test_property_car_name_split_on_null(self):
        instance = ParticipantPacket(self.binary_data(
            car_class_name=self.expected_car_name + "\x00Garbage Data"))
        expected_result = self.expected_car_name
        self.assertEqual(instance.car_name, expected_result)

    def test_property_data_hash(self):
        instance = ParticipantPacket(self.binary_data())
        expected_result = md5(self.binary_data()).hexdigest()
        self.assertEqual(instance.data_hash, expected_result)

    def test_property_name(self):
        instance = ParticipantPacket(self.binary_data())
        expected_result = self.expected_name
        self.assertListEqual(instance.name, expected_result)

    def test_property_name_split_on_null(self):
        instance = ParticipantPacket(self.binary_data(
            name=[name+'\x00Garbage Data' for name in self.expected_name]))
        expected_result = self.expected_name
        self.assertListEqual(instance.name, expected_result)

    def test_property_packet_type(self):
        instance = ParticipantPacket(self.binary_data())
        expected_result = self.expected_packet_type
        self.assertEqual(instance.packet_type, expected_result)

    def test_property_track_location(self):
        instance = ParticipantPacket(self.binary_data())
        expected_result = self.expected_track_location
        self.assertEqual(instance.track_location, expected_result)

    def test_property_track_location_split_on_null(self):
        instance = ParticipantPacket(self.binary_data(
            track_location=self.expected_track_location + "\x00Garbage Data"))
        expected_result = self.expected_track_location
        self.assertEqual(instance.track_location, expected_result)

    def test_property_track_variation(self):
        instance = ParticipantPacket(self.binary_data())
        expected_result = self.expected_track_variation
        self.assertEqual(instance.track_variation, expected_result)

    def test_property_track_variation_split_on_null(self):
        instance = ParticipantPacket(self.binary_data(
            track_variation=self.expected_track_variation + "\x00Garbage Data"))
        expected_result = self.expected_track_variation
        self.assertEqual(instance.track_variation, expected_result)

    def test_method_repr(self):
        instance = ParticipantPacket(self.binary_data())
        expected_result = "ParticipantPacket"
        self.assertEqual(repr(instance), expected_result)

    def test_method_str(self):
        instance = ParticipantPacket(self.binary_data())
        expected_result = "ParticipantPacket"
        self.assertEqual(str(instance), expected_result)



"""
Setup file for Replay Enhancer.
"""

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='replayenhancer',
    version='0.5.2',
    description='Replay Enhancer 0.5.2',
    long_description=long_description,
    url='https://github.com/SenorPez/project-cars-replay-enhancer',
    author='Senor Pez',
    author_email='contact_at_github@example.org',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords='gaming racing video data streaming',
    packages=find_packages(exclude=['assets', 'tests', 'utils']),
    install_requires=['moviepy', 'natsort', 'Pillow', 'tqdm'],
    extras_require={
        'dev': [],
        'test': []
    },
    package_data={
        'replayenhancer': [
            'lib/Roboto-Medium.ttf',
            'lib/Roboto-Regular.ttf',
            'lib/track_data.json']
    },
    entry_points={
        'console_scripts': [
            'replayenhancer=replayenhancer.replayenhancer:main',
            'packetcapture=replayenhancer.packetcapture:main'
        ],
    },
)




"""
Main execution script.
"""
import argparse
import json
import os
import sys

import moviepy.editor as mpy
from PIL import Image, ImageDraw
from moviepy.video.io.bindings import PIL_to_npimage
from tqdm import tqdm

from replayenhancer.DefaultCards \
    import SeriesChampion, SeriesStandings, StartingGrid
from replayenhancer.GTStandings import GTStandings
from replayenhancer.RaceData import RaceData
from replayenhancer.RaceResultsWithChange import RaceResultsWithChange
from replayenhancer.SeriesStandingsWithChange \
    import SeriesStandingsWithChange


def make_video(config_file, *, framerate=None, sync=False):
    configuration = json.load(open(config_file))
    try:
        race_data = RaceData(configuration['source_telemetry'])
        result_data = RaceData(configuration['source_telemetry'])
    except KeyError:
        sys.exit("Configuration Error: Source Telemetry not found.")

    try:
        output_prefix = os.path.splitext(configuration['output_video'])[0]
    except KeyError:
        sys.exit("Configuration Error: Output Video not found.")

    try:
        champion = configuration['show_champion']
    except KeyError:
        champion = False

    if os.environ.get('HEADINGFONTOVERRIDE') is not None:
        configuration['heading_font'] = \
            os.environ['HEADINGFONTOVERRIDE']
    if os.environ.get('DISPLAYFONTOVERRIDE') is not None:
        configuration['font'] = os.environ['DISPLAYFONTOVERRIDE']

    try:
        video_skipstart = configuration['video_skipstart']
    except KeyError:
        video_skipstart = 0

    try:
        video_skipend = configuration['video_skipend']
    except KeyError:
        video_skipend = None

    if 'source_video' in configuration \
            and configuration['source_video'] is not None:
        source_video = mpy.VideoFileClip(
            configuration['source_video']
        ).subclip(
            video_skipstart,
            video_skipend)
        if framerate is None:
            framerate = source_video.fps
    else:
        time_data = RaceData(configuration['source_telemetry'])
        with tqdm(desc="Detecting Telemetry Duration") as progress:
            while True:
                try:
                    _ = time_data.get_data()
                    progress.update()
                except StopIteration:
                    break
        source_video = mpy.ColorClip((1280, 1024)).set_duration(
            time_data.elapsed_time)

        if framerate is None:
            framerate = 30

    pcre_standings = GTStandings(
        race_data,
        ups=framerate,
        **configuration)
    standings_clip_mask = mpy.VideoClip(
        make_frame=pcre_standings.make_mask_frame,
        ismask=True)
    standings_clip = mpy.VideoClip(
        make_frame=pcre_standings.make_frame
    ).set_mask(standings_clip_mask)

    if sync:
        def timecode_frame(time):
            """
            Custom make frame for timecode.
            """
            timecode_image = Image.new('RGB', (100, 40))
            draw = ImageDraw.Draw(timecode_image)
            draw.text((10, 10), "%.02f"%(time))
            return PIL_to_npimage(timecode_image)

        timecode_clip = mpy.VideoClip(
            timecode_frame,
            duration=source_video.duration
        ).set_position(('center', 'top'))

        first_lap_data = RaceData(configuration['source_telemetry'])
        with tqdm(desc="Detecting Video Start") as progress:
            while not any(
                    [x.laps_complete > 0
                     for x in first_lap_data.drivers.values()]):
                _ = first_lap_data.get_data()
                progress.update()

        start_time = first_lap_data.elapsed_time - 10
        end_time = None

        with tqdm(desc="Detecting Video End") as progress:
            while not all(
                    [x.laps_complete > 0
                     for x in first_lap_data.drivers.values()]):
                try:
                    _ = first_lap_data.get_data()
                    progress.update()
                except StopIteration:
                    end_time = start_time + 60
                    break

        if end_time is None:
            end_time = first_lap_data.elapsed_time + 10

        main_event = mpy.CompositeVideoClip(
            [source_video, standings_clip, timecode_clip]
        ).set_duration(
            source_video.duration
        ).subclip(start_time, end_time)

    else:
        main_event = mpy.CompositeVideoClip(
            [source_video, standings_clip]
        ).set_duration(source_video.duration)

    pcre_starting_grid = StartingGrid(
        sorted(race_data.starting_grid, key=lambda x: x.position),
        size=source_video.size,
        **configuration)
    Image.fromarray(pcre_starting_grid.to_frame()).save(
        output_prefix + "_starting_grid.png")
    starting_grid = mpy.ImageClip(
        pcre_starting_grid.to_frame()).set_duration(5)

    while True:
        try:
            result_data.get_data()
        except StopIteration:
            break

    end_titles = list()

    pcre_results = RaceResultsWithChange(
        sorted(result_data.all_driver_classification, key=lambda x: x.position),
        result_data.starting_grid,
        size=source_video.size,
        **configuration)
    Image.fromarray(pcre_results.to_frame()).save(
        output_prefix + '_results.png')
    results = mpy.ImageClip(pcre_results.to_frame()).set_duration(20)

    end_titles.append(results)

    try:
        if any(configuration['point_structure']):
            if not any([
                    x['points'] for x
                    in configuration['participant_config'].values()]):
                pcre_series_standings = SeriesStandings(
                    result_data.all_driver_classification,
                    size=source_video.size,
                    **configuration)

                Image.fromarray(pcre_series_standings.to_frame()).save(
                    output_prefix + '_series_standings.png')
                series_standings = mpy.ImageClip(
                    pcre_series_standings.to_frame()).set_duration(20)

                end_titles.append(series_standings)
            else:
                pcre_series_standings = SeriesStandingsWithChange(
                    result_data.all_driver_classification,
                    size=source_video.size,
                    **configuration)

                Image.fromarray(pcre_series_standings.to_frame()).save(
                    output_prefix + '_series_standings.png')
                series_standings = mpy.ImageClip(
                    pcre_series_standings.to_frame()).set_duration(20)

                end_titles.append(series_standings)
    except KeyError:
        try:
            _ = configuration['point_structure']
            pcre_series_standings = SeriesStandings(
                result_data.all_driver_classification,
                size=source_video.size,
                **configuration)

            Image.fromarray(pcre_series_standings.to_frame()).save(
                output_prefix + '_series_standings.png')
            series_standings = mpy.ImageClip(
                pcre_series_standings.to_frame()).set_duration(20)

            end_titles.append(series_standings)
        except:
            pass

    if champion:
        pcre_series_champion = SeriesChampion(
            result_data.all_driver_classification,
            size=source_video.size,
            **configuration)
        Image.fromarray(pcre_series_champion.to_frame()).save(
            output_prefix + '_series_champion.png')
        series_champion = mpy.ImageClip(
            pcre_series_champion.to_frame()).set_duration(20)

        end_titles.append(series_champion)

    output = mpy.concatenate_videoclips([starting_grid.fadeout(1), main_event] + [clip.fadein(1).fadeout(1) for clip in end_titles[:-1]] + [end_titles[-1].fadein(1)], method="compose")

    output.write_videofile(configuration['output_video'], fps=float(framerate))


def main():
    parser = argparse.ArgumentParser(
        description="Project CARS Replay Enhancer")

    parser.add_argument('config')

    parser.add_argument(
        '-f',
        '--framerate',
        action='store',
        default=None)

    parser.add_argument(
        '-s',
        '--sync',
        action='store_true')

    parser.add_argument(
        '-v',
        '--version',
        action='version',
        version='Version 0.5.2')

    args = parser.parse_args()

    make_video(
        args.config,
        framerate=args.framerate,
        sync=args.sync)

if __name__ == '__main__':
    main()
